# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Callable, Dict, Iterator, List, Union

from kgforge.core import Resource
from kgforge.core.commons.attributes import repr_class
from kgforge.core.commons.execution import dispatch
from kgforge.core.conversions.json import as_json


class Reshaper:

    # This class is not intended to be specialized.

    def __init__(self, versioned_id_template: str) -> None:
        self.versioned_id_template: str = versioned_id_template

    def __repr__(self) -> str:
        return repr_class(self)

    def reshape(self, data: Union[Resource, List[Resource]], keep: List[str],
                versioned: bool) -> Union[Resource, List[Resource]]:
        # POLICY Resource _last_action and _store_metadata should be None.
        # POLICY Resource _validated and _synchronized should be False.
        return dispatch(data, self._reshape_many, self._reshape_one, keep, versioned)

    def _reshape_many(self, resources: List[Resource], keep: List[str],
                      versioned: bool) -> List[Resource]:
        # Could be optimized in the future.
        return [self._reshape(x, keep, versioned) for x in resources]

    def _reshape_one(self, resource: Resource, keep: List[str], versioned: bool) -> Resource:
        return self._reshape(resource, keep, versioned)

    def _reshape(self, resource: Resource, keep: List[str], versioned: bool) -> Resource:
        # TODO Use as base an implementation of JSONPath for Python. DKE-147.
        levels = [x.split(".", maxsplit=1) for x in keep]
        roots = {x[0] for x in levels}
        new = Resource()
        for root in roots:
            leafs = [x[1] for x in levels if len(x) > 1 and x[0] == root]
            value = getattr(resource, root)
            if isinstance(value, List):
                new_value = self._reshape_many(value, leafs, versioned)
            elif isinstance(value, Resource):
                if leafs:
                    new_value = self._reshape_one(value, leafs, versioned)
                else:
                    attributes = value.__dict__.items()
                    properties = {k: v for k, v in attributes if k not in value._RESERVED}
                    new_value = Resource(**properties)
            else:
                if root == "id" and versioned:
                    new_value = self.versioned_id_template.format(x=resource)
                else:
                    new_value = value
            setattr(new, root, new_value)
        return new


# TODO Use an implementation of JSONPath for Python instead to get values. DKE-147.
def collect_values(data: Union[Resource, List[Resource]],  follow: str,
                   exception: Callable = Exception) -> List[str]:
    def _collect(things: List) -> Iterator[str]:
        for x in things:
            if isinstance(x, Dict):
                for k, v in x.items():
                    if isinstance(v, List):
                        yield from _collect(v)
                    elif isinstance(v, Dict):
                        yield from _collect([v])
                    else:
                        yield v
    try:
        r = Reshaper("")
        reshaped = dispatch(data, r._reshape_many, r._reshape_one, [follow], False)
        jsoned = as_json(reshaped, False, False, None, None, None)
        prepared = jsoned if isinstance(jsoned, List) else [jsoned]
        return list(_collect(prepared))
    except AttributeError:
        raise exception("path to follow is incorrect")



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for conversion of a resource to / from triples.
import os
from urllib.parse import urljoin
from urllib.request import pathname2url

import json
import pytest
from rdflib import Graph, BNode

from kgforge.core.commons.exceptions import NotSupportedError
from kgforge.core.conversions.rdf import _merge_jsonld, from_jsonld, as_jsonld, Form, as_graph

form_store_metadata_combinations = [
    pytest.param(Form.COMPACTED.value, True, id="compacted-with-metadata"),
    pytest.param(Form.COMPACTED.value, False, id="compacted-without-metadata"),
    pytest.param(Form.EXPANDED.value, True, id="expanded-with-metadata"),
    pytest.param(Form.EXPANDED.value, False, id="expanded-without-metadata")
]

store_metadata_params = [
    pytest.param(True, id="with-metadata"),
    pytest.param(False, id="without-metadata")
]


@pytest.fixture
def building_with_context(building, model_context):
    building.context = model_context.document["@context"]
    return building


class TestJsonLd:

    @pytest.mark.parametrize("store_metadata", store_metadata_params)
    def test_nested_resources(self, organization, organization_jsonld_compacted,
                              store_metadata, model_context, metadata_context):
        compacted = organization_jsonld_compacted(organization, store_metadata=store_metadata)
        result = as_jsonld(organization, form=Form.COMPACTED.value, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context,
                           context_resolver=None)
        assert compacted == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_no_context(self, form, store_metadata, valid_resource):
        with pytest.raises(NotSupportedError):
            as_jsonld(valid_resource, form=form, store_metadata=store_metadata, model_context=None,
                      metadata_context=None, context_resolver=None)

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_unregistered_resource_with_context(self, building_with_context, building_jsonld,
                                                form, store_metadata, metadata_context):
        expected = building_jsonld(building_with_context, form, store_metadata, None)

        result = as_jsonld(building_with_context, form=form, store_metadata=store_metadata,
                           model_context=None, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_registered_resource_with_context(self, building_with_context, make_registered,
                                              building_jsonld, form, store_metadata,
                                              store_metadata_value, metadata_context):
        base = urljoin('file:', pathname2url(os.getcwd()))
        registered_resource = make_registered(building_with_context, store_metadata_value, base)
        expected = building_jsonld(registered_resource, form, store_metadata, None)
        result = as_jsonld(registered_resource, form=form, store_metadata=store_metadata,
                           model_context=None, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_unregistered_resource_model_context(self, building, model_context, building_jsonld,
                                                 form, store_metadata, metadata_context):
        expected = building_jsonld(building, form, store_metadata, model_context.document["@context"])
        result = as_jsonld(building, form=form, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context,
                           context_resolver=None)

        assert expected == result

    @pytest.mark.parametrize("form, store_metadata", form_store_metadata_combinations)
    def test_registered_resource_model_context(self, building_with_context, model_context,
                                               make_registered, building_jsonld, form,
                                               store_metadata, store_metadata_value,
                                               metadata_context):
        base = urljoin('file:', pathname2url(os.getcwd()))
        registered_resource = make_registered(building_with_context, store_metadata_value, base)
        expected = building_jsonld(registered_resource, form, store_metadata, None)
        result = as_jsonld(registered_resource, form=form, store_metadata=store_metadata,
                           model_context=model_context, metadata_context=metadata_context,
                           context_resolver=None)
        assert expected == result

    def test_from_jsonld(self, building, model_context, building_jsonld):
        building.context = model_context.document["@context"]
        payload = building_jsonld(building, "compacted", False, None)
        resource = from_jsonld(payload)
        assert resource == building

    def test_from_jsonld_2(self, building, model_context, building_jsonld):
        payload = {
            "@context": {
                "sh": "http://www.w3.org/ns/shacl#",
                "ex": "http://test/vocab#",
                "parent": {
                    "@id": "ex:parent",
                    "@type": "@id"
                },
                "nodeKind": {
                    "@id": "sh:nodeKind",
                    "@type": "@id"
                },
                "BlankNode": {
                    "@id": "sh:BlankNode"
                },
                "NodeShape": {
                    "@id": "sh:NodeShape"
                },
                "SuperClass": {
                    "@id": "ex:SuperClass"
                }
            },
            "@id": "http://test/data/123",
            "@type": "NodeShape",
            "parent": "ex:SuperClass",
            "nodeKind": "sh:BlankNode"
        }
        resource_dict = from_jsonld(payload)
        print(resource_dict)


        assert True

    def test_unresolvable_context(self, building, building_jsonld):
        building.context = "http://unresolvable.context.example.org/"
        payload = building_jsonld(building, "compacted", False, None)
        with pytest.raises(ValueError):
            from_jsonld(payload)


class TestGraph:

    def test_as_graph(self, building, building_jsonld, model_context):
        store_metadata = False
        building.id = "http://test/1234"
        building.context = model_context .document["@context"]
        data = building_jsonld(building, "compacted", store_metadata, None)
        expected = Graph().parse(data=json.dumps(data), format="json-ld")
        result = as_graph(building, store_metadata, model_context, None, None)
        for s, p, o in expected:
            if isinstance(o, BNode):
                assert (s, p, None) in result
            elif isinstance(s, BNode):
                assert (None, p, o) in result
            else:
                assert (s, p, o) in result


class TestUtils:

    def test_merge_jsonld(self):
        str_ctx = "string_context"
        str_ctx2 = "string_context2"
        dict_ctx = {"a": 1, "b": 2}
        dict_ctx2 = {"c": 3, "d": 4}
        list_ctx = [str_ctx, dict_ctx]
        list_ctx2 = [str_ctx2, dict_ctx2]
        dict_ctx3 = {"a": 1, "b": 2, "c": 3, "d": 4}
        assert _merge_jsonld(str_ctx, str_ctx2) == [str_ctx, str_ctx2]
        assert _merge_jsonld(str_ctx, dict_ctx) == [str_ctx, dict_ctx]
        assert _merge_jsonld(str_ctx2, list_ctx) == [str_ctx2, str_ctx, dict_ctx]
        assert _merge_jsonld(list_ctx, str_ctx2) == [str_ctx, dict_ctx, str_ctx2]
        assert _merge_jsonld(list_ctx, dict_ctx2) == [str_ctx, dict_ctx3]
        assert _merge_jsonld(list_ctx, list_ctx2) == [str_ctx, dict_ctx3, str_ctx2]
        assert _merge_jsonld(dict_ctx, str_ctx) == [dict_ctx, str_ctx]
        assert _merge_jsonld(dict_ctx, list_ctx2) == [dict_ctx3, str_ctx2]
        assert _merge_jsonld(dict_ctx, dict_ctx2) == dict_ctx3
        assert _merge_jsonld(str_ctx, str_ctx) == str_ctx
        assert _merge_jsonld(dict_ctx, dict_ctx) == dict_ctx
        assert _merge_jsonld(list_ctx, list_ctx) == list_ctx
        assert _merge_jsonld(list_ctx, None) == list_ctx



 #
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from pytest_bdd import given, parsers, scenarios, then, when

from kgforge.specializations.stores.demo_store import DemoStore
from tests.conftest import check_report, do

# TODO To be port to the generic parameterizable test suite for stores in test_stores.py. DKE-135.


scenarios("demo_store.feature")


@given("A store instance.")
def store():
    return DemoStore()


@given("An already registered resource.", target_fixture="data")
def registered_resource(store, valid_resource):
    store.register(valid_resource)
    assert valid_resource._synchronized is True
    return valid_resource


@given("Already registered resources.", target_fixture="data")
def registered_resources(store, valid_resources):
    store.register(valid_resources)
    for x in valid_resources:
        assert x._synchronized is True
        assert x._store_metadata == {'version': 1, 'deprecated': False}
    return valid_resources


@when(parsers.re("I register the resource(?P<rc>s?)."
                 " The printed report does(?P<err> not)? mention an error(: '(?P<msg>[a-zA-Z0-9: ]+)')?."))
def register(capsys, store, data, rc, err, msg):
    store.register(data)
    check_report(capsys, rc, err, msg, "_register_one")


@when("I register the resource. An exception is raised. The printed report does mention an error: 'Exception: exception raised'.")
def register_exception(monkeypatch, capsys, store, data):
    def _register_one(_, x, schema_id): raise Exception("exception raised")
    monkeypatch.setattr("kgforge.specializations.stores.demo_store.DemoStore._register_one", _register_one)
    store.register(data)
    out = capsys.readouterr().out[:-1]
    assert out == f"<action> _register_one\n<succeeded> False\n<error> Exception: exception raised"


@then(parsers.parse("The store metadata of a resource should be '{metadata}'."))
def check_metadata(data, metadata):
    def fun(x): assert str(x._store_metadata) == metadata
    do(fun, data)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import os
from unittest import mock
from urllib.parse import urljoin
from urllib.request import pathname2url
from uuid import uuid4

import nexussdk
import pytest
from typing import Callable, Union, List

from kgforge.core import Resource
from kgforge.core.archetypes import Store
from kgforge.core.commons.context import Context
from kgforge.core.conversions.rdf import _merge_jsonld
from kgforge.core.wrappings.dict import wrap_dict
from kgforge.core.wrappings.paths import Filter
from kgforge.specializations.stores.bluebrain_nexus import BlueBrainNexus, build_query_statements

# FIXME mock Nexus for unittests
# TODO To be port to the generic parameterizable test suite for stores in test_stores.py. DKE-135.
from kgforge.specializations.stores.nexus.service import NEXUS_CONTEXT

BUCKET = "test/kgforge"
NEXUS = "https://nexus-instance.org/"
TOKEN = "token"
NEXUS_PROJECT_CONTEXT = {"base": "http://data.net/", "vocab": "http://vocab.net/"}

VERSIONED_TEMPLATE = "{x.id}?rev={x._store_metadata._rev}"
FILE_RESOURCE_MAPPING = os.sep.join((os.path.curdir, "tests", "data", "nexus-store",
                                     "file-to-resource-mapping.hjson"))


@pytest.fixture
def nested_resource():
    contributions = [Resource(title=f"contribution {i}") for i in range(3)]
    return Resource(type="Agent", name="someone", contributions=contributions)


@pytest.fixture
def nested_registered_resource(nested_resource):
    ingredients = [Resource(id=i, type='Ingredient') for i in range(3)]
    resource = Resource(id="a_recipe", type="Recipe", ingridients=ingredients,
                        author=Resource(id="a_person", type="Person"))
    do_recursive(add_metadata, resource)
    return resource


@pytest.fixture
def metadata_data_compacted():
    return {
        "_deprecated": False,
        "_updatedBy": "http://integration.kfgorge.test",
        "_rev": 1
    }


@pytest.fixture
def store_metadata_value(metadata_data_compacted):
    data = {"id": "placeholder"}
    data.update(metadata_data_compacted)
    return data


@pytest.fixture
def registered_building(building, model_context, store_metadata_value):
    building.context = model_context.iri if model_context.is_http_iri() else model_context.document[
        "@context"]
    if model_context.base:
        building.id = f"{model_context.base}{str(uuid4())}"
    else:
        building.id = f"{urljoin('file:', pathname2url(os.getcwd()))}/{str(uuid4())}"
    store_metadata_value["id"] = building.id
    building._store_metadata = wrap_dict(store_metadata_value)
    return building


@pytest.fixture
def registered_person(person, store_metadata_value):
    custom_context = person.context
    if custom_context.base:
        person.id = f"{person.base}{str(uuid4())}"
    else:
        person.id = f"{urljoin('file:', pathname2url(os.getcwd()))}/{str(uuid4())}"
    store_metadata_value["id"] = person.id
    person._store_metadata = wrap_dict(store_metadata_value)
    return person


@pytest.fixture
@mock.patch('nexussdk.projects.fetch', return_value=NEXUS_PROJECT_CONTEXT)
@mock.patch('nexussdk.resources.fetch', side_effect=nexussdk.HTTPError('404'))
def nexus_store(context_project_patch, metadata_context_patch):
    return BlueBrainNexus(endpoint=NEXUS, bucket=BUCKET, token=TOKEN,
                          file_resource_mapping=FILE_RESOURCE_MAPPING)


@pytest.fixture
def nexus_store_unauthorized():
    return BlueBrainNexus(endpoint=NEXUS, bucket=BUCKET, token="invalid token")


def test_config_error():
    with pytest.raises(ValueError):
        BlueBrainNexus(endpoint="test", bucket="invalid", token="")


def test_config(nexus_store):
    assert nexus_store.organisation == "test"
    assert nexus_store.project == "kgforge"
    assert nexus_store.endpoint == NEXUS
    assert nexus_store.context.base == NEXUS_PROJECT_CONTEXT["base"]


def test_freeze_fail(nexus_store: Store, nested_resource):
    """nested resource is not registered, thus freeze will fail"""
    nexus_store.versioned_id_template = "{x.id}?rev={x._store_metadata._rev}"
    nested_resource.id = "abc"
    add_metadata(nested_resource)


def test_freeze_nested(nexus_store: Store, nested_registered_resource):
    nexus_store.versioned_id_template = "{x.id}?rev={x._store_metadata._rev}"
    nexus_store.freeze(nested_registered_resource)
    do_recursive(assert_frozen_id, nested_registered_resource)


def test_to_resource(nexus_store, registered_building, building_jsonld):
    context = _merge_jsonld(registered_building.context, NEXUS_CONTEXT)
    payload = building_jsonld(registered_building, "compacted", True, None)
    payload["@context"] = context
    result = nexus_store.service.to_resource(payload)
    assert str(result) == str(registered_building)
    assert getattr(result, "context") == registered_building.context
    assert str(result._store_metadata) == str(registered_building._store_metadata)


class TestQuerying:

    @pytest.fixture
    def context(self):
        document = {
            "@context": {
                "contribution": {
                    "@id": "https://neuroshapes.org/contribution",
                    "@type": "@id"
                },
                "agent": {
                    "@id": "http://www.w3.org/ns/prov#agent",
                    "@type": "@id"
                },
                "Person": "http://schema.org/Person",
                "address": "http://schema.org/address",
                "name": "http://schema.org/name",
                "postalCode": "http://schema.org/postalCode",
                "streetAddress": "http://schema.org/streetAddress",
                "deprecated": "https://bluebrain.github.io/nexus/vocabulary/deprecated"
            }
        }
        return Context(document)

    @pytest.mark.parametrize("filters,expected", [
        pytest.param((Filter(["agent", "name"], "__eq__", "Allen Institute"),),
                     (["agent/name ?v0"], ["FILTER(?v0 = \"Allen Institute\")"]),
                     id="literal"),
        pytest.param((Filter(["address", "postalCode"], "__lt__", 50070),),
                     (["address/postalCode ?v0"], ["FILTER(?v0 < 50070)"]),
                     id="number-lt"),
        pytest.param((Filter(["address", "postalCode"], "__gt__", 50070),),
                     (["address/postalCode ?v0"], ["FILTER(?v0 > 50070)"]),
                     id="number-gt"),
        pytest.param((Filter(["address", "postalCode"], "__ge__", 50070),),
                     (["address/postalCode ?v0"], ["FILTER(?v0 >= 50070)"]),
                     id="number-ge"),
        pytest.param((Filter(["address", "postalCode"], "__le__", 50070),),
                     (["address/postalCode ?v0"], ["FILTER(?v0 <= 50070)"]),
                     id="number-le"),
        pytest.param((Filter(["deprecated"], "__eq__", False),),
                     (["deprecated ?v0"], ["FILTER(?v0 = false)"]),
                     id="boolean-false"),
        pytest.param((Filter(["deprecated"], "__eq__", True),),
                     (["deprecated ?v0"], ["FILTER(?v0 = true)"]),
                     id="boolean-true"),
        pytest.param((Filter(["type"], "__eq__", "Person"),),
                     (["type Person"], []),
                     id="iri-eq"),
        pytest.param((Filter(["type"], "__ne__", "Person"),),
                     (["type ?v0"], ["FILTER(?v0 != Person)"]),
                     id="iri-ne"),
        pytest.param((Filter(["type"], "__ne__", "Person"), Filter(["name"], "__eq__", "toto")),
                     (["type ?v0", "name ?v1"],
                      ["FILTER(?v0 != Person)", "FILTER(?v1 = \"toto\")"]),
                     id="iri-ne-name-eq")
    ])
    def test_filter_to_query_statements(self, context, filters, expected):
        statements = build_query_statements(context, filters)
        assert statements == expected


# Helpers


def assert_frozen_id(resource: Resource):
    assert resource.id.endswith('?rev=' + str(resource._store_metadata['_rev']))


def add_metadata(resource: Resource):
    metadata = {
        "_self": resource.id,
        "_constrainedBy": "https://bluebrain.github.io/nexus/schemas/unconstrained.json",
        "_project": "https://nexus/org/prj",
        "_rev": 1,
        "_deprecated": False,
        "_createdAt": "2019-03-28T13:40:38.934Z",
        "_createdBy": "https://nexus/u1",
        "_updatedAt": "2019-03-28T13:40:38.934Z",
        "_updatedBy": "https://nexus/u1",
        "_incoming": "https:/nexus/incoming",
        "_outgoing": "https://nexux/outgoing"
    }
    resource._synchronized = True
    resource._validated = True
    resource._store_metadata = wrap_dict(metadata)


def do_recursive(fun: Callable, data: Union[Resource, List[Resource]], *args) -> None:
    if isinstance(data, List) and all(isinstance(x, Resource) for x in data):
        for x in data:
            fun(x, *args)
    elif isinstance(data, Resource):
        fun(data, *args)
        for _, v in data.__dict__.items():
            if isinstance(v, (Resource, List)):
                do_recursive(fun, v, *args)
    else:
        raise TypeError("not a Resource nor a list of Resource")



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Placeholder for the test suite for actions.

from kgforge.core import Resource
from kgforge.core.commons.actions import LazyAction, collect_lazy_actions, execute_lazy_actions


def test_execute_lazy_actions():
    fun = lambda x: x
    ra = Resource(pa1="pa1",
                  pa2=Resource(pb1="pb1"),
                  pa3=LazyAction(fun, "pa3 executed"),
                  pa4=Resource(pc1=LazyAction(fun, "pc1 executed"), pc2="pc2"),
                  pa5=[LazyAction(fun, "pa5[0] executed"),
                       123,
                       Resource(pd1=LazyAction(fun, "pd1 executed")),
                       "string"])
    la = collect_lazy_actions(ra)
    execute_lazy_actions(ra, la)
    assert ra.pa1 == "pa1"
    assert ra.pa2.pb1 == "pb1"
    assert ra.pa3 == "pa3 executed"
    assert ra.pa4.pc1 == "pc1 executed"
    assert ra.pa4.pc2 == "pc2"
    assert ra.pa5[0] == "pa5[0] executed"
    assert ra.pa5[1] == 123
    assert ra.pa5[2].pd1 == "pd1 executed"
    assert ra.pa5[3] == "string"



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import yaml
from pandas import DataFrame
from rdflib import Graph

from kgforge.core import Resource
from kgforge.core.archetypes import Mapping, Model, Resolver, Store
from kgforge.core.commons.actions import LazyAction
from kgforge.core.commons.dictionaries import with_defaults
from kgforge.core.commons.exceptions import ResolvingError
from kgforge.core.commons.execution import catch
from kgforge.core.commons.imports import import_class
from kgforge.core.commons.strategies import ResolvingStrategy
from kgforge.core.conversions.dataframe import as_dataframe, from_dataframe
from kgforge.core.conversions.json import as_json, from_json
from kgforge.core.conversions.rdf import as_jsonld, from_jsonld, as_graph, from_graph, Form
from kgforge.core.reshaping import Reshaper
from kgforge.core.wrappings.paths import PathsWrapper, wrap_paths
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping


class KnowledgeGraphForge:

    # POLICY Class name should be imported in the corresponding module __init__.py.

    # No catching of exceptions so that no incomplete instance is created if an error occurs.
    # This is a best practice in Python for __init__().
    def __init__(self, configuration: Union[str, Dict], **kwargs) -> None:

        # Required minimal configuration: name for Model and Store, origin and source for Model.
        # Keyword arguments could be used to override the configuration provided for the Store.
        #
        # The configuration could be provided either as:
        #
        #   - A path to a YAML file with the following structure.
        #     Class name could be provided in three formats:
        #       * 'SomeClass',
        #       * 'SomeClass from package.module',
        #       * 'SomeClass from package.module.file'.
        #     When the class is from this package, the first is used. Otherwise, the two others.
        #     Values using braces with something inside should be quoted with double quotes.
        #     See demo-forge.yml in kgforge/examples/configurations/ for an example.
        #
        # Model:
        #   name: <a class name of a Model>
        #   origin: <'directory', 'url', or 'store'>
        #   source: <a directory path, an URL, or the class name of a Store>
        #   bucket: <when 'origin' is 'store', a Store bucket>
        #   endpoint: <when 'origin' is 'store', a Store endpoint, default to Store:endpoint>
        #   token: <when 'origin' is 'store', a Store token, default to Store:token>
        #   context:
        #     iri: <an IRI>
        #     bucket: <when 'origin' is 'store', a Store bucket, default to Model:bucket>
        #     endpoint: <when 'origin' is 'store', a Store endpoint, default to Model:endpoint>
        #     token: <when 'origin' is 'store', a Store token, default to Model:token>
        #
        # Store:
        #   name: <a class name of a Store>
        #   endpoint: <an URL>
        #   bucket: <a bucket as a string>
        #   token: <a token as a string>
        #   versioned_id_template: <a string template using 'x' to access resource fields>
        #   file_resource_mapping: <an Hjson string, a file path, or an URL>
        #
        # Resolvers:
        #   <scope>:
        #     - resolver: <a class name of a Resolver>
        #       origin: <'directory', 'web_service', or 'store'>
        #       source: <a directory path, a web service endpoint, or the class name of a Store>
        #       targets:
        #         - identifier: <a name, or an IRI>
        #           bucket: <a file name, an URL path, or a Store bucket>
        #       result_resource_mapping: <an Hjson string, a file path, or an URL>
        #       endpoint: <when 'origin' is 'store', a Store endpoint, default to Store:endpoint>
        #       token: <when 'origin' is 'store', a Store token, default to Store:token>
        #
        # Formatters:
        #   <identifier>: <a string template with replacement fields delimited by braces, i.e. '{}'>
        #
        #   - A Python dictionary with the following structure.
        #
        # {
        #     "Model": {
        #         "name": <str>,
        #         "origin": <str>,
        #         "source": <str>,
        #         "bucket": <str>,
        #         "endpoint": <str>,
        #         "token": <str>,
        #         "context": {
        #               "iri": <str>,
        #               "bucket": <str>,
        #               "endpoint": <str>,
        #               "token": <str>,
        #         }
        #     },
        #     "Store": {
        #         "name": <str>,
        #         "endpoint": <str>,
        #         "bucket": <str>,
        #         "token": <str>,
        #         "versioned_id_template": <str>,
        #         "file_resource_mapping": <str>,
        #     },
        #     "Resolvers": {
        #         "<scope>": [
        #             {
        #                 "resolver": <str>,
        #                 "origin": <str>,
        #                 "source": <str>,
        #                 "targets": [
        #                     {
        #                         "identifier": <str>,
        #                         "bucket": <str>,
        #                     },
        #                     ...,
        #                 ],
        #                 "result_resource_mapping": <str>,
        #                 "endpoint": <str>,
        #                 "token": <str>,
        #             },
        #             ...,
        #         ],
        #     },
        #     "Formatters": {
        #         "<name>": <str>,
        #         ...,
        #     },
        # }
        if isinstance(configuration, str):
            with Path(configuration).open(encoding="utf-8") as f:
                config = yaml.safe_load(f)
        else:
            config = deepcopy(configuration)

        # Debugging.

        self._debug = kwargs.pop("debug", False)

        # Store.

        store_config = config.pop("Store")
        store_config.update(kwargs)

        # Model.

        model_config = config.pop("Model")
        if model_config["origin"] == "store":
            with_defaults(model_config, store_config, "source", "name", ["endpoint", "token", "bucket"])
        model_name = model_config.pop("name")
        model = import_class(model_name, "models")
        self._model: Model = model(**model_config)

        # Store.

        store_config.update(model_context=self._model.context())
        store_name = store_config.pop("name")
        store = import_class(store_name, "stores")
        self._store: Store = store(**store_config)
        store_config.update(name=store_name)

        # Resolvers.

        resolvers_config = config.pop("Resolvers", None)
        # Format: Optional[Dict[scope_name, Dict[resolver_name, Resolver]]].
        self._resolvers: Optional[Dict[str, Dict[str, Resolver]]] = prepare_resolvers(
            resolvers_config, store_config) if resolvers_config else None

        # Formatters.

        self._formatters: Optional[Dict[str, str]] = config.pop("Formatters", None)

    # Modeling User Interface.

    @catch
    def prefixes(self, pretty: bool = True) -> Optional[Dict[str, str]]:
        return self._model.prefixes(pretty)

    @catch
    def types(self, pretty: bool = True) -> Optional[List[str]]:
        return self._model.types(pretty)

    @catch
    def template(self, type: str, only_required: bool = False, output: str = "hjson"
                 ) -> Optional[Dict]:
        return self._model.template(type, only_required, output)

    # No @catch because the error handling is done by execution.run().
    def validate(self, data: Union[Resource, List[Resource]],
                 execute_actions_before: bool = False) -> None:
        self._model.validate(data, execute_actions_before)

    # Resolving User Interface.

    def resolvers(self):
        print("Available scopes:")
        for k, v in sorted(self._resolvers.items()):
            print(" - ", k, ":")
            for r_key, r_value in v.items():
                print("     - resolver: ", r_key)
                print("         - targets: ",  ",".join(r_value.targets.keys()))

    @catch
    def resolve(self, text: str, scope: Optional[str] = None, resolver: Optional[str] = None,
                target: Optional[str] = None, type: Optional[str] = None,
                strategy: ResolvingStrategy = ResolvingStrategy.BEST_MATCH,
                limit: Optional[int] = 10) -> Optional[Union[Resource, List[Resource]]]:
        """
        Resolve a string into a existing resource or list of resources depending on the resolving
        strategy

        Args:
            text (str): string to search.
            scope (str): a scope identifier
            resolver (str): a resolver name
            target (str): an identifier in the targets
            type (str): a Type to be used as part of the query
            strategy (str): a ResolvingStrategy.[ALL_MATCHES, BEST_MATCH, EXACT_MATCH]
            limit (int): when using ALL_MATCHES provide a limited number of results, default is 10
        Returns:
            resources (Resource, List(Resource))
        """
        if self._resolvers is not None:
            if scope is not None:
                if resolver is not None:
                    rov = self._resolvers[scope][resolver]
                else:
                    scope_resolvers = list(self._resolvers[scope].values())
                    if len(scope_resolvers) == 1:
                        rov = scope_resolvers[0]
                    else:
                        raise ResolvingError("resolver name should be specified")
            else:
                scopes_resolvers = list(self._resolvers.values())
                if resolver is not None:
                    if len(self._resolvers) == 1:
                        rov = scopes_resolvers[0][resolver]
                    else:
                        raise ResolvingError("resolving scope should be specified")
                else:
                    if len(self._resolvers) == 1 and len(scopes_resolvers[0]) == 1:
                        rov = list(scopes_resolvers[0].values())[0]
                    else:
                        raise ResolvingError("resolving scope or resolver name should be "
                                             "specified")
            if type and self._model.context():
                type = self._model.context().expand(type)
            return rov.resolve(text, target, type, strategy, limit)
        else:
            raise ResolvingError("no resolvers have been configured")

    # Formatting User Interface.

    @catch
    def format(self, what: str, *args) -> str:
        return self._formatters[what].format(*args)

    # Mapping User Interface.

    @catch
    def sources(self, pretty: bool = True) -> Optional[List[str]]:
        return self._model.sources(pretty)

    @catch
    def mappings(self, source: str, pretty: bool = True) -> Optional[Dict[str, List[str]]]:
        return self._model.mappings(source, pretty)

    @catch
    def mapping(self, entity: str, source: str, type: Callable = DictionaryMapping) -> Mapping:
        return self._model.mapping(entity, source, type)

    @catch
    def map(self, data: Any, mapping: Union[Mapping, List[Mapping]],
            mapper: Callable = DictionaryMapper, na: Union[Any, List[Any]] = None
            ) -> Union[Resource, List[Resource]]:
        return mapper(self).map(data, mapping, na)

    # Reshaping User Interface.

    @catch
    def reshape(self, data: Union[Resource, List[Resource]], keep: List[str],
                versioned: bool = False) -> Union[Resource, List[Resource]]:
        return Reshaper(self._store.versioned_id_template).reshape(data, keep, versioned)

    # Querying User Interface.

    @catch
    def retrieve(self, id: str, version: Optional[Union[int, str]] = None,
                 cross_bucket: bool = False) -> Resource:
        return self._store.retrieve(id, version, cross_bucket)

    @catch
    def paths(self, type: str) -> PathsWrapper:
        template = self._model.template(type, False, "dict")
        return wrap_paths(template)

    @catch
    def search(self, *filters, **params) -> List[Resource]:
        resolvers = list(self._resolvers.values()) if self._resolvers is not None else None
        return self._store.search(resolvers, *filters, **params)

    @catch
    def sparql(self, query: str, debug: bool = False, limit: int = 100,
               offset: Optional[int] = None) -> List[Resource]:
        return self._store.sparql(query, debug, limit, offset)

    @catch
    def download(self, data: Union[Resource, List[Resource]], follow: str, path: str,
                 overwrite: bool = False) -> None:
        # path: DirPath.
        self._store.download(data, follow, path, overwrite)

    # Storing User Interface.

    # No @catch because the error handling is done by execution.run().
    def register(self, data: Union[Resource, List[Resource]], schema_id: Optional[str]=None) -> None:
        self._store.register(data, schema_id)

    # No @catch because the error handling is done by execution.run().
    def update(self, data: Union[Resource, List[Resource]]) -> None:
        self._store.update(data)

    # No @catch because the error handling is done by execution.run().
    def deprecate(self, data: Union[Resource, List[Resource]]) -> None:
        self._store.deprecate(data)

    # Versioning User Interface.

    # No @catch because the error handling is done by execution.run().
    def tag(self, data: Union[Resource, List[Resource]], value: str) -> None:
        self._store.tag(data, value)

    # No @catch because the error handling is done by execution.run().
    def freeze(self, data: Union[Resource, List[Resource]]) -> None:
        self._store.freeze(data)

    # Files Handling User Interface.

    @catch
    def attach(self, path: str, content_type: str = None) -> LazyAction:
        # path: Union[FilePath, DirPath].
        return LazyAction(self._store.upload, path, content_type)

    # Converting User Interface.

    @catch
    def as_json(self, data: Union[Resource, List[Resource]], expanded: bool = False,
                store_metadata: bool = False) -> Union[Dict, List[Dict]]:
        return as_json(data, expanded, store_metadata, self._model.context(),
                       self._store.metadata_context, self._model.resolve_context)

    @catch
    def as_jsonld(self, data: Union[Resource, List[Resource]], form: str = Form.COMPACTED.value,
                  store_metadata: bool = False) -> Union[Dict, List[Dict]]:
        return as_jsonld(data, form, store_metadata, self._model.context(),
                         self._store.metadata_context, self._model.resolve_context)

    @catch
    def as_graph(self, data: Union[Resource, List[Resource]], store_metadata: bool) -> Graph:
        return as_graph(data, store_metadata, self._model.context(), self._store.metadata_context,
                        self._model.resolve_context)

    @catch
    def as_dataframe(self, data: List[Resource], na: Union[Any, List[Any]] = [None], nesting: str = ".",
                     expanded: bool = False, store_metadata: bool = False) -> DataFrame:
        return as_dataframe(data, na, nesting, expanded, store_metadata, self._model.context(),
                            self._store.metadata_context, self._model.resolve_context)

    @catch
    def from_json(self, data: Union[Dict, List[Dict]], na: Union[Any, List[Any]] = None
                  ) -> Union[Resource, List[Resource]]:
        return from_json(data, na)

    @catch
    def from_jsonld(self, data: Union[Dict, List[Dict]]) -> Union[Resource, List[Resource]]:
        return from_jsonld(data)

    @catch
    def from_graph(self, data: Graph) -> Union[Resource, List[Resource]]:
        return from_graph(data)

    @catch
    def from_dataframe(self, data: DataFrame, na: Union[Any, List[Any]] = np.nan,
                       nesting: str = ".") -> Union[Resource, List[Resource]]:
        return from_dataframe(data, na, nesting)


def prepare_resolvers(config: Dict, store_config: Dict) -> Dict[str, Dict[str, Resolver]]:
    return {scope: dict(prepare_resolver(x, store_config) for x in configs)
            for scope, configs in config.items()}


def prepare_resolver(config: Dict, store_config: Dict) -> Tuple[str, Resolver]:
    if config["origin"] == "store":
        with_defaults(config, store_config, "source", "name", ["endpoint", "token", "bucket", "model_context"])
    resolver_name = config.pop("resolver")
    resolver = import_class(resolver_name, "resolvers")
    return resolver.__name__, resolver(**config)



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from pytest_bdd import given, parsers, scenarios, when

from kgforge.specializations.models.demo_model import DemoModel
from tests.conftest import check_report

# TODO To be port to the generic parameterizable test suite for models in test_models.py. DKE-135.


scenarios("demo_model.feature")


@given("A model instance.")
def model():
    return DemoModel("tests/data/demo-model/", origin="directory")


@given("A validated resource.", target_fixture="data")
def validated_resource(model, valid_resource):
    model.validate(valid_resource, execute_actions_before=False)
    assert valid_resource._validated == True
    return valid_resource


@when(parsers.re("I validate the resource(?P<rc>s?)."
                 " The printed report does(?P<err> not)? mention an error(: '(?P<msg>[a-zA-Z0-9: ]+)')?."))
def validate(capsys, model, data, rc, err, msg):
    model.validate(data, execute_actions_before=False)
    check_report(capsys, rc, err, msg, "_validate_one")


@when("I validate the resource. An exception is raised. The printed report mentions an error: 'Exception: exception raised'.")
def validate_exception(monkeypatch, capsys, model, data):
    def _validate_one(_, x): raise Exception("exception raised")
    monkeypatch.setattr("kgforge.specializations.models.demo_model.DemoModel._validate_one", _validate_one)
    model.validate(data, execute_actions_before=False)
    out = capsys.readouterr().out[:-1]
    assert out == f"<action> _validate_one\n<succeeded> False\n<error> Exception: exception raised"



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

# Test suite for initializing a forge.

import pytest

from kgforge.core import KnowledgeGraphForge

SCOPE = "terms"
MODEL = "DemoModel"
STORE = "DemoStore"
RESOLVER = "DemoResolver"


@pytest.fixture(params=[
    MODEL,
    f"{MODEL} from kgforge.specializations.models",
    f"{MODEL} from kgforge.specializations.models.demo_model"])
def model(request):
    return request.param


@pytest.fixture(params=[
    STORE,
    f"{STORE} from kgforge.specializations.stores",
    f"{STORE} from kgforge.specializations.stores.demo_store"])
def store(request):
    return request.param


@pytest.fixture(params=[
    RESOLVER,
    f"{RESOLVER} from kgforge.specializations.resolvers",
    f"{RESOLVER} from kgforge.specializations.resolvers.demo_resolver"])
def resolver(request):
    return request.param


@pytest.fixture
def config(model, store, resolver):
    return {
        "Model": {
            "name": model,
            "origin": "directory",
            "source": "tests/data/demo-model/",
        },
        "Store": {
            "name": store,
        },
        "Resolvers": {
            SCOPE: [
                {
                    "resolver": resolver,
                    "origin": "directory",
                    "source": "tests/data/demo-resolver/",
                    "targets": [
                        {
                            "identifier": "sex",
                            "bucket": "sex.json",
                        },
                    ],
                    "result_resource_mapping": "../../configurations/demo-resolver/term-to-resource-mapping.hjson",
                },
            ],
        },
    }


class TestForgeInitialization:

    def test_initialization(self, config):
        forge = KnowledgeGraphForge(config)
        assert type(forge._model).__name__ == MODEL
        assert type(forge._store).__name__ == STORE
        assert type(forge._resolvers[SCOPE][RESOLVER]).__name__ == RESOLVER



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from copy import deepcopy
from typing import Dict, List, Optional, Union
from uuid import uuid4

from kgforge.core import Resource
from kgforge.core.archetypes import Resolver, Store
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import (DeprecationError, RegistrationError,
                                             RetrievalError, TaggingError, UpdatingError)
from kgforge.core.commons.execution import not_supported
from kgforge.core.conversions.json import as_json, from_json
from kgforge.core.wrappings.dict import wrap_dict


class DemoStore(Store):
    """An example to show how to implement a Store and to demonstrate how it is used."""

    def __init__(self, endpoint: Optional[str] = None, bucket: Optional[str] = None,
                 token: Optional[str] = None, versioned_id_template: Optional[str] = None,
                 file_resource_mapping: Optional[str] = None,
                 model_context: Optional[Context] = None) -> None:
        super().__init__(endpoint, bucket, token, versioned_id_template, file_resource_mapping,
                         model_context)

    # [C]RUD.

    def _register_one(self, resource: Resource, schema_id: str) -> None:
        data = as_json(resource, expanded=False, store_metadata=False, model_context=None,
                       metadata_context=None, context_resolver=None)
        try:
            record = self.service.create(data)
        except StoreLibrary.RecordExists:
            raise RegistrationError("resource already exists")
        else:
            resource.id = record["data"]["id"]
            resource._store_metadata = wrap_dict(record["metadata"])

    # C[R]UD.

    def retrieve(self, id: str, version: Optional[Union[int, str]],
                 cross_bucket: bool) -> Resource:
        if cross_bucket:
            not_supported(("cross_bucket", True))
        try:
            record = self.service.read(id, version)
        except StoreLibrary.RecordMissing:
            raise RetrievalError("resource not found")
        else:
            return _to_resource(record)

    # CR[U]D.

    def _update_one(self, resource: Resource) -> None:
        data = as_json(resource, expanded=False, store_metadata=False, model_context=None,
                       metadata_context=None, context_resolver=None)
        try:
            record = self.service.update(data)
        except StoreLibrary.RecordMissing:
            raise UpdatingError("resource not found")
        except StoreLibrary.RecordDeprecated:
            raise UpdatingError("resource is deprecated")
        else:
            resource._store_metadata = wrap_dict(record["metadata"])

    def _tag_one(self, resource: Resource, value: str) -> None:
        # Chosen case: tagging does not modify the resource.
        rid = resource.id
        version = resource._store_metadata.version
        try:
            self.service.tag(rid, version, value)
        except StoreLibrary.TagExists:
            raise TaggingError("resource version already tagged")
        except StoreLibrary.RecordMissing:
            raise TaggingError("resource not found")

    # CRU[D].

    def _deprecate_one(self, resource: Resource) -> None:
        rid = resource.id
        try:
            record = self.service.deprecate(rid)
        except StoreLibrary.RecordMissing:
            raise DeprecationError("resource not found")
        except StoreLibrary.RecordDeprecated:
            raise DeprecationError("resource already deprecated")
        else:
            resource._store_metadata = wrap_dict(record["metadata"])

    # Querying.

    def search(self, resolvers: Optional[List[Resolver]], *filters, **params) -> List[Resource]:
        # TODO DKE-145.
        print("<info> DemoStore does not support handling of errors with QueryingError for now.")
        # TODO DKE-145.
        print("<info> DemoStore does not support traversing lists for now.")
        if params:
            # TODO DKE-145.
            print("DemoStore does not support 'resolving' and 'lookup' parameters for now.")
        conditions = [f"x.{'.'.join(x.path)}.{x.operator}({x.value!r})" for x in filters]
        records = self.service.find(conditions)
        return [_to_resource(x) for x in records]

    # Utils.

    def _initialize_service(self, endpoint: Optional[str], bucket: Optional[str],
                            token: Optional[str]):
        return StoreLibrary()


def _to_resource(record: Dict) -> Resource:
    # TODO This operation might be abstracted in core when other stores will be implemented.
    resource = from_json(record["data"], None)
    resource._store_metadata = wrap_dict(record["metadata"])
    resource._synchronized = True
    return resource


class StoreLibrary:
    """Simulate a third-party library handling interactions with the database used by the store."""

    def __init__(self):
        self.records: Dict[str, Dict] = {}
        self.archives: Dict[str, Dict] = {}
        self.tags: Dict[str, int] = {}

    def create(self, data: Dict) -> Dict:
        record = self._record(data, 1, False)
        rid = record["data"]["id"]
        if rid in self.records.keys():
            raise self.RecordExists
        self.records[rid] = record
        return record

    def read(self, rid: str, version: Optional[Union[int, str]]) -> Dict:
        try:
            if version is not None:
                if isinstance(version, str):
                    tkey = self._tag_id(rid, version)
                    version = self.tags[tkey]
                akey = self._archive_id(rid, version)
                record = self.archives[akey]
            else:
                record = self.records[rid]
        except KeyError:
            raise self.RecordMissing
        else:
            return record

    def update(self, data: Dict) -> Dict:
        rid = data.get("id", None)
        try:
            record = self.records[rid]
        except KeyError:
            raise self.RecordMissing
        else:
            metadata = record["metadata"]
            if metadata["deprecated"]:
                raise self.RecordDeprecated
            version = metadata["version"]
            key = self._archive_id(rid, version)
            self.archives[key] = record
            new_record = self._record(data, version + 1, False)
            self.records[rid] = new_record
            return new_record

    def deprecate(self, rid: str) -> Dict:
        try:
            record = self.records[rid]
        except KeyError:
            raise self.RecordMissing
        else:
            metadata = record["metadata"]
            if metadata["deprecated"]:
                raise self.RecordDeprecated
            version = metadata["version"]
            key = self._archive_id(rid, version)
            self.archives[key] = record
            data = record["data"]
            new_record = self._record(data, version + 1, True)
            self.records[rid] = new_record
            return new_record

    def tag(self, rid: str, version: int, value: str) -> None:
        if rid in self.records:
            key = self._tag_id(rid, value)
            if key in self.tags:
                raise self.TagExists
            else:
                self.tags[key] = version
        else:
            raise self.RecordMissing

    def find(self, conditions: List[str]) -> List[Dict]:
        return [r for r in self.records.values()
                if all(eval(c, {}, {"x": wrap_dict(r["data"])}) for c in conditions)]

    def _record(self, data: Dict, version: int, deprecated: bool) -> Dict:
        copy = deepcopy(data)
        if "id" not in copy:
            copy["id"] = self._new_id()
        return {
            "data": copy,
            "metadata": {
                "version": version,
                "deprecated": deprecated,
            },
        }

    @staticmethod
    def _new_id() -> str:
        return str(uuid4())

    @staticmethod
    def _archive_id(rid: str, version: int) -> str:
        return f"{rid}_version={version}"

    @staticmethod
    def _tag_id(rid: str, tag: str) -> str:
        return f"{rid}_tag={tag}"

    class RecordExists(Exception):
        pass

    class RecordMissing(Exception):
        pass

    class RecordDeprecated(Exception):
        pass

    class TagExists(Exception):
        pass



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import os

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file.
with open(os.path.join(HERE, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="nexusforge",
    author="Blue Brain Project, EPFL",
    use_scm_version={
        "write_to": "kgforge/version.py",
        "write_to_template": "__version__ = '{version}'\n",
    },
    description="Building and Using Knowledge Graphs made easy.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords="framework knowledge graph data science",
    url="https://github.com/BlueBrain/nexus-forge",
    packages=find_packages(),
    python_requires=">=3.6",
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[
        "hjson",
        "pyyaml",
        "pandas",
        "nexus-sdk",
        "aiohttp",
        "nest_asyncio",
        "pyLD",
        "pyshacl==0.11.6.post1",
        "rdflib-jsonld"
    ],
    extras_require={
        "dev": ["tox", "pytest", "pytest-bdd", "pytest-cov", "pytest-mock", "codecov"],
        "docs": ["sphinx", "sphinx-bluebrain-theme"],
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3 :: Only",
        "Natural Language :: English",
    ]
)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import inspect
import traceback
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple, Union

from kgforge.core import Resource
from kgforge.core.commons.actions import (Action, Actions, collect_lazy_actions,
                                          execute_lazy_actions)
from kgforge.core.commons.exceptions import NotSupportedError


# POLICY Should have only one function called 'wrapper'. See catch().


def not_supported(arg: Optional[Tuple[str, Any]] = None) -> None:
    # TODO When 'arg' is specified, compare with the value in the frame to know if it applies.
    # POLICY Should be called in methods in core which could be not implemented by specializations.
    frame = inspect.currentframe().f_back
    try:
        self_ = frame.f_locals["self"]
        class_name = type(self_).__name__
        method_name = inspect.getframeinfo(frame).function
        tail = f" with {arg[0]}={arg[1]}" if arg else ""
        msg = f"{class_name} is not supporting {method_name}(){tail}"
    except Exception as e:
        raise e
    else:
        raise NotSupportedError(msg)
    finally:
        del frame


def catch(fun):
    # POLICY Should decorate methods operating on Resource instances.
    # POLICY If recovering from errors is needed, use execution.run().
    # NB: The class of the methods should have only one KnowledgeGraphForge instance as attribute.

    @wraps(fun)
    def wrapper(*args, **kwargs):

        class_name = "KnowledgeGraphForge"
        self = args[0]
        if type(self).__name__ == class_name:
            forge = self
        else:
            forge = next(x for x in self.__dict__.values() if type(x).__name__ == class_name)
        debug = forge._debug

        try:
            return fun(*args, **kwargs)
        except Exception as e:
            stack = traceback.extract_stack()
            it = (x for x in stack if x.name == "wrapper" and x.filename == stack[-1].filename)
            next(it)
            try:
                next(it)
            except StopIteration:
                called_once = True
            else:
                called_once = False
            finally:
                if called_once and not debug:
                    tb = e.__traceback__
                    fs = traceback.extract_tb(tb)[-1]
                    print(f"<action> {fs.name}"
                          f"\n<error> {type(e).__name__}: {e}\n")
                    return None
                else:
                    raise

    return wrapper


# @functools.singledispatchmethod is introduced in Python 3.8.
def dispatch(data: Union[Resource, List[Resource]], fun_many: Callable,
             fun_one: Callable, *args) -> Any:
    # POLICY The method calling this function should be decorated with execution.catch().
    if isinstance(data, List) and all(isinstance(x, Resource) for x in data):
        return fun_many(data, *args)
    elif isinstance(data, Resource):
        return fun_one(data, *args)
    else:
        raise TypeError("not a Resource nor a list of Resource")


def run(fun_one: Callable, fun_many: Optional[Callable], data: Union[Resource, List[Resource]],
        exception: Callable, id_required: bool = False,
        required_synchronized: Optional[bool] = None, execute_actions: bool = False,
        monitored_status: Optional[str] = None, catch_exceptions: bool = True, **kwargs) -> None:
    # POLICY Should be called for operations on resources where recovering from errors is needed.
    if isinstance(data, List) and all(isinstance(x, Resource) for x in data):
        if fun_many is None:
            _run_many(fun_one, data, exception, id_required, required_synchronized,
                      execute_actions, monitored_status, catch_exceptions, **kwargs)
        else:
            fun_many(data, **kwargs)
        actions = Actions.from_resources(data)
        print(actions)
    elif isinstance(data, Resource):
        _run_one(fun_one, data, exception, id_required, required_synchronized, execute_actions,
                 monitored_status, catch_exceptions, **kwargs)
        action = data._last_action
        print(action)
    else:
        raise TypeError("not a Resource nor a list of Resource")


def _run_many(fun: Callable, resources: List[Resource], *args, **kwargs) -> None:
    for x in resources:
        _run_one(fun, x, *args, **kwargs)


def _run_one(fun: Callable, resource: Resource, exception: Callable, id_required: bool,
             required_synchronized: Optional[bool], execute_actions: bool,
             monitored_status: Optional[str], catch_exceptions: bool, **kwargs) -> None:
    try:
        if id_required and not hasattr(resource, "id"):
            raise exception("resource should have an id")

        synchronized = resource._synchronized
        if required_synchronized is not None and synchronized is not required_synchronized:
            be_or_not_be = "be" if required_synchronized is True else "not be"
            raise exception(f"resource should {be_or_not_be} synchronized")

        lazy_actions = collect_lazy_actions(resource)
        if execute_actions:
            execute_lazy_actions(resource, lazy_actions)
        elif lazy_actions:
            raise exception("resource has lazy actions which need to be executed before")

        result = fun(resource, **kwargs)
    except Exception as e:
        status = False
        succeeded = False
        exception = e
    else:
        status = True if not isinstance(result, bool) else result
        succeeded = True
        exception = None
    finally:
        if monitored_status:
            setattr(resource, monitored_status, status)

        resource._last_action = Action(fun.__name__, succeeded, exception)

        if not catch_exceptions and exception:
            raise exception



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import json
import pytest

from kgforge.core import Resource
from kgforge.core.commons.exceptions import ValidationError
from kgforge.specializations.models import RdfModel
from tests.specializations.models.data import *


@pytest.fixture
def rdf_model(context_iri_file):
    return RdfModel("tests/data/shacl-model",
                    context={"iri": context_iri_file},
                    origin="directory")


class TestVocabulary:

    def test_generate_context(self, rdf_model: RdfModel):
        generated_context = rdf_model._generate_context()
        for k in TYPES_SCHEMAS_MAP.keys():
            assert generated_context.expand(k) is not None

    def test_types(self, rdf_model: RdfModel):
        types = rdf_model.types(pretty=False)
        assert types == list(TYPES_SCHEMAS_MAP.keys())

    def test_context(self, rdf_model: RdfModel, context_file_path):
        with open(context_file_path) as f:
            expected = json.load(f)
        vocabulary = rdf_model.context().document
        assert vocabulary == expected

    def test_namespaces(self, rdf_model: RdfModel, model_prefixes):
        assert rdf_model.prefixes(pretty=False) == model_prefixes


class TestTemplates:

    def test_request_invalid_type(self, rdf_model: RdfModel):
        with pytest.raises(ValueError):
            rdf_model._template("Invalid", False)

    @pytest.mark.parametrize("type_, expected", [
        pytest.param("Person", PERSON_TEMPLATE, id="person"),
        pytest.param("Employee", EMPLOYEE_TEMPLATE, id="employee"),
        pytest.param("Activity", ACTIVITY_TEMPLATE, id="activity"),
        pytest.param("Building", BUILDING_TEMPLATE, id="building"),
    ])
    def test_create_templates(self, rdf_model: RdfModel, type_, expected):
        result = rdf_model._template(type_, False)
        assert result == expected

    @pytest.mark.parametrize("type_, expected", [
        pytest.param("Activity", ACTIVITY_TEMPLATE_MANDATORY, id="activity"),
        pytest.param("Building", BUILDING_TEMPLATE_MANDATORY, id="building"),
    ])
    def test_create_templates_only_required(self, rdf_model: RdfModel, type_, expected):
        result = rdf_model._template(type_, True)
        assert result == expected


class TestValidation:

    @pytest.fixture
    def entity_resource(self):
        return Resource(type="Entity")

    @pytest.fixture
    def activity_json(self, entity_resource):
        return {"type": "Activity", "generated": entity_resource, "status": "completed"}

    @pytest.fixture
    def invalid_activity_resource(self, activity_json):
        return Resource(**activity_json)

    @pytest.fixture
    def valid_activity_resource(self, activity_json):
        resource = Resource(**activity_json)
        resource.id = "http://testing/123"
        return resource

    @pytest.mark.parametrize("type_,", TYPES_SCHEMAS_MAP.keys())
    def test_type_to_schema(self, rdf_model: RdfModel, type_):
        # FIXME TYPES_SCHEMAS_MAP should be a type to file dictionary
        assert rdf_model.schema_id(type_) == TYPES_SCHEMAS_MAP[type_]

    def test_validate_one(self, rdf_model: RdfModel, valid_activity_resource):
        rdf_model.validate(valid_activity_resource, False)

    def test_validate_one_fail(self, rdf_model: RdfModel, invalid_activity_resource):
        with pytest.raises(ValidationError):
            rdf_model._validate_one(invalid_activity_resource)

    def test_validate_many(self, rdf_model: RdfModel, valid_activity_resource,
                           invalid_activity_resource):
        resources = [valid_activity_resource,
                     invalid_activity_resource]
        rdf_model.validate(resources, False)
        assert valid_activity_resource._validated is True
        assert invalid_activity_resource._validated is False
        assert (valid_activity_resource._last_action.operation ==
                invalid_activity_resource._last_action.operation ==
                rdf_model._validate_many.__name__)



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from urllib.parse import urlparse


def as_term(value: str) -> str:
    try:
        return split_uri(value)[1]
    except AttributeError:
        return value


def split_uri(uri):
    parsed = urlparse(uri)
    if parsed.fragment:
        fragment = parsed.fragment
    else:
        path_split = parsed.path.split("/")
        fragment = path_split[len(path_split)-1]
    return uri.replace(fragment, ""), fragment



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
from pandas import DataFrame, Series

from kgforge.core import Resource
from kgforge.core.commons.context import Context
from kgforge.core.conversions.json import as_json, from_json


def as_dataframe(data: List[Resource], na: Union[Any, List[Any]], nesting: str, expanded: bool,
                 store_metadata: bool, model_context: Optional[Context],
                 metadata_context: Optional[Context], context_resolver: Optional[Callable]) -> DataFrame:
    dicts = as_json(data, expanded, store_metadata, model_context=model_context,
                    metadata_context=metadata_context, context_resolver=context_resolver)
    # NB: Do not use json_normalize(). It does not respect how the dictionaries are ordered.
    flattened = (flatten(x, nesting) for x in dicts)
    df = DataFrame(flattened)
    if na is not None:
        df.replace(na, np.nan, inplace=True)
    return df


def flatten(data: Dict, sep: str) -> Dict:
    return dict(_flatten(data, sep, []))


def _flatten(data: Dict, sep: str, path: List[str]) -> Iterator[Tuple[str, Any]]:
    for k, v in data.items():
        p = [*path, k]
        if isinstance(v, Dict):
            yield from _flatten(v, sep, p)
        else:
            yield sep.join(p), v


def from_dataframe(data: DataFrame, na: Union[Any, List[Any]], nesting: str
                   ) -> Union[Resource, List[Resource]]:
    converted = [_from_dataframe(x, na, nesting) for _, x in data.iterrows()]
    if len(converted) == 1:
        converted = converted[0]
    return converted


def _from_dataframe(row: Series, na: Union[Any, List[Any]], nesting: str) -> Resource:
    new_na = row.replace(na, np.nan)
    no_na = new_na.dropna()
    items = list(no_na.items())
    data = deflatten(items, nesting)
    return from_json(data, None)


def deflatten(items: List[Tuple[str, Any]], sep: str) -> Dict:
    d = dict()
    i = 0
    while i < len(items):
        k, v = items[i]
        if sep not in k:
            d[k] = v
            i += 1
        else:
            pk, _, ck = k.partition(sep)
            pitems = list(take_while(items[i:], f"{pk}{sep}"))
            d[pk] = deflatten(pitems, sep)
            i += len(pitems)
    return d


def take_while(items: List[Tuple[str, Any]], start: str) -> Iterator[Tuple[str, Any]]:
    for k, v in items:
        if k.startswith(start):
            yield k.replace(start, "", 1), v
        else:
            break



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
from typing import Dict, Optional, Union, List, Tuple

import json

from pyshacl import Validator

from kgforge.core import Resource
from kgforge.core.commons.exceptions import RetrievalError
from kgforge.core.conversions.rdf import as_jsonld, as_graph
from rdflib import URIRef, Namespace, Graph

from kgforge.core.archetypes import Store
from kgforge.specializations.models.rdf.node_properties import NodeProperties
from kgforge.specializations.models.rdf.service import RdfService, ShapesGraphWrapper


class StoreService(RdfService):

    def __init__(self, default_store: Store, context_iri: Optional[str] = None,
                 context_store: Optional[Store] = None) -> None:

        self.default_store = default_store
        self.context_store = context_store or default_store
        self.NXV = Namespace("https://bluebrain.github.io/nexus/vocabulary/")
        self.store_metadata_iri = "https://bluebrain.github.io/nexus/contexts/resource.json"
        self._shapes_to_resources: Dict
        self._imported = list()
        self._graph = Graph()
        self._sg = ShapesGraphWrapper(self._graph)
        super().__init__(self._graph, context_iri)

    def schema_source_id(self, schema_iri: str) -> str:
        return self._shapes_to_resources[schema_iri]

    def materialize(self, iri: URIRef) -> NodeProperties:
        shape = self._type_shape(iri)
        predecessors = set()
        props, attrs = shape.traverse(predecessors)
        if props:
            attrs["properties"] = props
        return NodeProperties(**attrs)

    def _validate(self, iri: str, data_graph: Graph) -> Tuple[bool, Graph, str]:
        # _type_shape will make sure all the shapes for this type are in the graph
        self._type_shape(iri)
        validator = Validator(data_graph, shacl_graph=self._graph)
        return validator.run()

    def resolve_context(self, iri: str) -> Dict:
        if iri in self._context_cache:
            return self._context_cache[iri]
        document = self.recursive_resolve(iri)
        self._context_cache.update({iri: document})
        return document

    def generate_context(self) -> Dict:
        for v in self._shapes_to_resources.values():
            self._load_shape(v)
        # reloads the shapes graph
        self._sg = ShapesGraphWrapper(self._graph)
        return self._generate_context()

    def _build_shapes_map(self) -> Dict:
        query = f"""
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX sh: <http://www.w3.org/ns/shacl#>
            SELECT ?type ?shape ?resource_id WHERE {{
                {{ ?shape sh:targetClass ?type .
                   ?resource_id <{self.NXV.shapes}> ?shape
                }} UNION {{
                    SELECT (?shape as ?type) ?shape ?resource_id WHERE {{
                        ?shape a sh:NodeShape .
                        ?shape a rdfs:Class .
                        ?resource_id <{self.NXV.shapes}> ?shape
                    }}
                }}
            }} ORDER BY ?type"""
        # make sure to get all types
        limit = 100
        offset = 0
        count = limit
        class_to_shapes = dict()
        shape_resource = dict()
        while count == limit:
            resources = self.context_store.sparql(query, debug=False, limit=limit, offset=offset)
            for r in resources:
                shape_uri = URIRef(r.shape)
                class_to_shapes[r.type] = shape_uri
                shape_resource[shape_uri] = URIRef(r.resource_id)
            count = len(resources)
            offset += limit
        self._shapes_to_resources = shape_resource
        return class_to_shapes

    def recursive_resolve(self, context: Union[Dict, List, str]) -> Dict:
        document = dict()
        if isinstance(context, list):
            if self.store_metadata_iri in context:
                context.remove(self.store_metadata_iri)
            for x in context:
                document.update(self.recursive_resolve(x))
        elif isinstance(context, str):
            try:
                local_only = False if self.default_store == self.context_store else True
                doc = self.default_store.service.resolve_context(context, local_only=local_only)
            except ValueError:
                try:
                    doc = self.context_store.service.resolve_context(context, local_only=False)
                except ValueError as e:
                    raise e
            document.update(self.recursive_resolve(doc))
        elif isinstance(context, dict):
            document.update(context)
        return document

    def _load_shape(self, resource_id):
        if resource_id not in self._imported:
            try:
                shape = self.context_store.retrieve(resource_id, version=None, cross_bucket=False)
            except RetrievalError as e:
                print(e, resource_id)
                # failed, don't try to load again
                self._imported.append(resource_id)
            else:
                json_dict = as_jsonld(shape, form="compacted", store_metadata=False, model_context=None,
                                      metadata_context=None,
                                      context_resolver=self.context_store.service.resolve_context)
                # this double conversion was due blank nodes were not "regenerated" with json-ld
                temp_graph = Graph().parse(data=json.dumps(json_dict), format="json-ld")
                self._graph.parse(data=temp_graph.serialize(format="n3"), format="n3")
                self._imported.append(resource_id)
                if hasattr(shape, "imports"):
                    for dependency in shape.imports:
                        self._load_shape(self.context.expand(dependency))

    def _type_shape(self, iri: URIRef):
        try:
            shape = self._sg.lookup_shape_from_node(iri)
        except KeyError:
            self._load_shape(self._shapes_to_resources[iri])
            # reloads the shapes graph
            self._sg = ShapesGraphWrapper(self._graph)
            shape = self._sg.lookup_shape_from_node(iri)
        return shape



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from kgforge.core import Resource
from kgforge.core.commons.attributes import repr_class
from kgforge.core.commons.exceptions import ConfigurationError
from kgforge.core.commons.execution import not_supported
from kgforge.core.commons.imports import import_class
from kgforge.core.commons.strategies import ResolvingStrategy


class Resolver(ABC):

    # See demo_resolver.py in kgforge/specializations/resolvers/ for a reference implementation.

    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/resolvers/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Create a generic parameterizable test suite for resolvers. DKE-135.
    # POLICY Implementations should pass tests/specializations/resolvers/test_resolvers.py.

    def __init__(self, source: str, targets: List[Dict[str, str]], result_resource_mapping: str,
                 **source_config) -> None:
        # POLICY Resolver data access should be lazy, unless it takes less than a second.
        # POLICY There could be data caching but it should be aware of changes made in the source.
        self.source: str = source
        self.targets: Dict[str, str] = {x["identifier"]: x["bucket"] for x in targets}
        self.result_mapping: Any = self.mapping.load(result_resource_mapping)
        self.service: Any = self._initialize_service(self.source, self.targets, **source_config)

    def __repr__(self) -> str:
        return repr_class(self)

    @property
    @abstractmethod
    def mapping(self) -> Callable:
        """Mapping class to load result_resource_mapping."""
        pass

    @property
    @abstractmethod
    def mapper(self) -> Callable:
        """Mapper class to map the result data to a Resource with result_resource_mapping."""
        pass

    def resolve(self, text: str, target: Optional[str], type: Optional[str],
                strategy: ResolvingStrategy, limit: Optional[int]) -> Optional[Union[Resource, List[Resource]]]:
        # The resolving strategy cannot be abstracted as it should be managed by the service.
        resolved = self._resolve(text, target, type, strategy, limit)
        if resolved is not None:
            if len(resolved) == 1:
                resolved = resolved[0]
            return self.mapper().map(resolved, self.result_mapping, None)
        else:
            return None

    @abstractmethod
    def _resolve(self, text: str, target: Optional[str], type: Optional[str],
                 strategy: ResolvingStrategy, limit: Optional[str]) -> Optional[List[Any]]:
        # POLICY Should notify of failures with exception ResolvingError including a message.
        pass

    # Utils.

    def _initialize_service(self, source: str, targets: Dict[str, str], **source_config) -> Any:
        # Resolver data could be accessed from a directory, a web service, or a Store.
        # Initialize the access to the resolver data according to the source type.
        # POLICY Should not use 'self'. This is not a function only for the specialization to work.
        origin = source_config.pop("origin")
        if origin == "directory":
            dirpath = Path(source)
            return self._service_from_directory(dirpath, targets)
        elif origin == "web_service":
            return self._service_from_web_service(source, targets)
        elif origin == "store":
            store = import_class(source, "stores")
            return self._service_from_store(store, targets, **source_config)
        else:
            raise ConfigurationError(f"unrecognized Resolver origin '{origin}'")

    @staticmethod
    @abstractmethod
    def _service_from_directory(dirpath: Path, targets: Dict[str, str]) -> Any:
        pass

    @staticmethod
    def _service_from_web_service(endpoint: str, targets: Dict[str, str]) -> Any:
        not_supported()

    @staticmethod
    def _service_from_store(store: Callable, targets: Dict[str, str], **store_config) -> Any:
        not_supported()



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Any, Dict, List, Union

from kgforge.core.commons.attributes import check_collisions, repr_class


class Filter:

    def __init__(self, path: List[str], operator: str, value: Any) -> None:
        self.path: List[str] = path
        self.operator: str = operator
        self.value: Any = value

    def __repr__(self) -> str:
        return repr_class(self)


class FilterMixin:

    _RESERVED = {"_path"}

    def __init__(self, path: List[str]) -> None:
        self._path: List[str] = path

    def __repr__(self) -> str:
        return repr_class(self)

    def __lt__(self, other: Any) -> Filter:
        return self._for("__lt__", other)

    def __le__(self, other: Any) -> Filter:
        return self._for("__le__", other)

    def __eq__(self, other: Any) -> Filter:
        return self._for("__eq__", other)

    def __ne__(self, other: Any) -> Filter:
        return self._for("__ne__", other)

    def __gt__(self, other: Any) -> Filter:
        return self._for("__gt__", other)

    def __ge__(self, other: Any) -> Filter:
        return self._for("__ge__", other)

    def _for(self, operator: str, other: Any) -> Filter:
        return Filter(self._path, operator, other)


class PathWrapper(FilterMixin):

    def __init__(self, path: List[str]) -> None:
        super().__init__(path)


class PathsWrapper(FilterMixin):

    def __init__(self, path: List[str], paths: Dict) -> None:
        check_collisions(self._RESERVED, paths.keys())
        super().__init__(path)
        self.__dict__ = paths


def wrap_paths(template: Dict) -> PathsWrapper:
    return _wrap(template, [])


def _wrap(data: Any, path: List[str]) -> Union[PathsWrapper, PathWrapper]:
    if isinstance(data, Dict):
        return PathsWrapper(path, {k: _wrap(v, path + [k]) for k, v in data.items()})
    else:
        return PathWrapper(path)



# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import re
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Match, Optional, Union

from kgforge.core import Resource
from kgforge.core.commons.attributes import repr_class
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import (DeprecationError, DownloadingError, FreezingError,
                                             RegistrationError, TaggingError, UpdatingError,
                                             UploadingError)
from kgforge.core.commons.execution import not_supported, run
from kgforge.core.reshaping import collect_values


# NB: Do not 'from kgforge.core.archetypes import Resolver' to avoid cyclic dependency.


class Store(ABC):

    # See demo_store.py in kgforge/specializations/stores/ for a reference implementation.

    # POLICY Methods of archetypes, except __init__, should not have optional arguments.

    # POLICY Implementations should be declared in kgforge/specializations/stores/__init__.py.
    # POLICY Implementations should not add methods but private functions in the file.
    # TODO Move from BDD to classical testing to have a more parameterizable test suite. DKE-135.
    # POLICY Implementations should pass tests/specializations/stores/demo_store.feature tests.

    def __init__(self, endpoint: Optional[str] = None, bucket: Optional[str] = None,
                 token: Optional[str] = None, versioned_id_template: Optional[str] = None,
                 file_resource_mapping: Optional[str] = None,
                 model_context: Optional[Context] = None) -> None:
        # file_resource_mapping: Optional[Union[Hjson, FilePath, URL]].
        # POLICY There could be data caching but it should be aware of changes made in the source.
        self.endpoint: Optional[str] = endpoint
        self.bucket: Optional[str] = bucket
        self.token: Optional[str] = token
        self.versioned_id_template: Optional[str] = versioned_id_template
        loaded = self.mapping.load(file_resource_mapping) if file_resource_mapping else None
        self.file_mapping: Optional[Any] = loaded
        self.model_context: Optional[Context] = model_context
        self.service: Any = self._initialize_service(self.endpoint, self.bucket, self.token)
        self.context: Context = self.service.context if hasattr(self.service, "context") else None
        self.metadata_context: Context = self.service.metadata_context if hasattr(self.service, "metadata_context") else None

    def __repr__(self) -> str:
        return repr_class(self)

    @property
    def mapping(self) -> Optional[Callable]:
        """Mapping class to load file_resource_mapping."""
        return None

    @property
    def mapper(self) -> Optional[Callable]:
        """Mapper class to map file metadata to a Resource with file_resource_mapping."""
        return None

    # [C]RUD.

    def register(self, data: Union[Resource, List[Resource]], schema_id: str = None) -> None:
        # Replace None by self._register_many to switch to optimized bulk registration.
        run(self._register_one, None, data, required_synchronized=False, execute_actions=True,
            exception=RegistrationError, monitored_status="_synchronized", schema_id=schema_id)

    def _register_many(self, resources: List[Resource], schema_id: str) -> None:
        # Bulk registration could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._register_one() and execution._run_one() behaviours.
        not_supported()

    @abstractmethod
    def _register_one(self, resource: Resource, schema_id: str) -> None:
        # POLICY Should notify of failures with exception RegistrationError including a message.
        # POLICY Resource _store_metadata should be set using wrappers.dict.wrap_dict().
        # TODO This operation might be abstracted here when other stores will be implemented.
        pass

    # This expected that '@catch' is not used here. This is for actions.execute_lazy_actions().
    def upload(self, path: str, content_type: str) -> Union[Resource, List[Resource]]:
        # path: Union[FilePath, DirPath].
        if self.file_mapping is not None:
            p = Path(path)
            uploaded = self._upload(p, content_type)
            return self.mapper().map(uploaded, self.file_mapping, None)
        else:
            raise UploadingError("no file_resource_mapping has been configured")

    def _upload(self, path: Path, content_type: str) -> Union[Any, List[Any]]:
        # path: Union[FilePath, DirPath].
        if path.is_dir():
            filepaths = [x for x in path.iterdir() if x.is_file() and not x.name.startswith(".")]
            return self._upload_many(filepaths, content_type)
        else:
            return self._upload_one(path, content_type)

    def _upload_many(self, paths: List[Path], content_type: str) -> List[Any]:
        # Bulk uploading could be optimized by overriding this method in the specialization.
        # POLICY Should follow self._upload_one() policies.
        return [self._upload_one(x, content_type) for x in paths]

    def _upload_one(self, path: Path, content_type: str) -> Any:
        # path: FilePath.
        # POLICY Should notify of failures with exception UploadingError including a message.
        not_supported()

    # C[R]UD.

    @abstractmethod
    def retrieve(self, id: str, version: Optional[Union[int, str]], cross_bucket: bool
                 ) -> Resource:
        # POLICY Should notify of failures with exception RetrievalError including a message.
        # POLICY Resource _store_metadata should be set using wrappers.dict.wrap_dict().
        # POLICY Resource _synchronized should be set to True.
        # TODO These two operations might be abstracted here when other stores will be implemented.
        pass

    def _retrieve_filename(self, id: str) -> str:
        # TODO This operation might be adapted if other file metadata are needed.
        not_supported()

    def download(self, data: Union[Resource, List[Resource]], follow: str, path: str,
                 overwrite: bool) -> None:
        # path: DirPath.
        urls = collect_values(data, follow, DownloadingError)
        count = len(urls)
        if count == 0:
            raise DownloadingError("no URLs were found")
        dirpath = Path(path)
        dirpath.mkdir(parents=True, exist_ok=True)
        timestamp = time.strftime("%Y%m%d%H%M%S")
        filepaths = []
        for x in urls:
            filename = self._retrieve_filename(x)
            filepath = dirpath / filename
            if not overwrite and filepath.exists():
                filepaths.append(f"{filepath}.{timestamp}")
            else:
                filepaths.append(str(filepath))
        if count > 1:
            self._download_many(urls, filepaths)
        else:
            self._download_one(urls[0], filepaths[0])

    def _download_many(self, urls: List[str], paths: List[str]) -> None:
        # paths: List[FilePath].
        # Bulk downloading could be optimized by overriding this method in the specialization.
        # POLICY Should follow self._download_one() policies.
        for url, path in zip(urls, paths):
            self._download_one(url, path)

    def _download_one(self, url: str, path: str) -> None:
        # path: FilePath.
        # POLICY Should notify of failures with exception DownloadingError including a message.
        not_supported()

    # CR[U]D.

    def update(self, data: Union[Resource, List[Resource]]) -> None:
        # Replace None by self._update_many to switch to optimized bulk update.
        run(self._update_one, None, data, id_required=True, required_synchronized=False,
            execute_actions=True, exception=UpdatingError, monitored_status="_synchronized")

    def _update_many(self, resources: List[Resource]) -> None:
        # Bulk update could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._update_one() and execution._run_one() behaviours.
        not_supported()

    @abstractmethod
    def _update_one(self, resource: Resource) -> None:
        # POLICY Should notify of failures with exception UpdatingError including a message.
        # POLICY Resource _store_metadata should be set using wrappers.dict.wrap_dict().
        # TODO This operation might be abstracted here when other stores will be implemented.
        pass

    def tag(self, data: Union[Resource, List[Resource]], value: str) -> None:
        # Replace None by self._tag_many to switch to optimized bulk tagging.
        # POLICY If tagging modify the resource, run() should have status='_synchronized'.
        run(self._tag_one, None, data, id_required=True, required_synchronized=True,
            exception=TaggingError, value=value)

    def _tag_many(self, resources: List[Resource], value: str) -> None:
        # Bulk tagging could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._tag_one() and execution._run_one() behaviours.
        # POLICY If tagging modify the resource, it should be done with status='_synchronized'.
        not_supported()

    def _tag_one(self, resource: Resource, value: str) -> None:
        # POLICY Should notify of failures with exception TaggingError including a message.
        # POLICY If tagging modify the resource, _store_metadata should be updated.
        not_supported()

    # CRU[D].

    def deprecate(self, data: Union[Resource, List[Resource]]) -> None:
        # Replace None by self._deprecate_many to switch to optimized bulk deprecation.
        run(self._deprecate_one, None, data, id_required=True, required_synchronized=True,
            exception=DeprecationError, monitored_status="_synchronized")

    def _deprecate_many(self, resources: List[Resource]) -> None:
        # Bulk deprecation could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._deprecate_one() and execution._run_one() behaviours.
        not_supported()

    def _deprecate_one(self, resource: Resource) -> None:
        # POLICY Should notify of failures with exception DeprecationError including a message.
        # POLICY Resource _store_metadata should be set using wrappers.dict.wrap_dict().
        # TODO This operation might be abstracted here when other stores will be implemented.
        not_supported()

    # Querying.

    def search(self, resolvers: Optional[List["Resolver"]], *filters, **params) -> List[Resource]:

        # Positional arguments in 'filters' are instances of type Filter from wrappings/paths.py.
        # Keyword arguments in 'params' could be:
        #   - debug: bool,
        #   - limit: int,
        #   - offset: int,
        #   - deprecated: bool,
        #   - resolving: str, with values in ('exact', 'fuzzy'),
        #   - lookup: str, with values in ('current', 'children').
        # POLICY Should use sparql() when SPARQL is chosen here has the querying language.
        # POLICY Should notify of failures with exception QueryingError including a message.
        # POLICY Resource _store_metadata should be set using wrappers.dict.wrap_dict().
        # POLICY Resource _synchronized should be set to True.
        # TODO These two operations might be abstracted here when other stores will be implemented.
        not_supported()

    def sparql(self, query: str, debug: bool, limit: int, offset: int = None) -> List[Resource]:
        if self.model_context is not None:
            context = self.model_context.document
            prefixes = self.model_context.prefixes
        else:
            context = None
            prefixes = None
        qr = rewrite_sparql(query, context, prefixes) if context is not None else query
        if debug:
            print(*["Submitted query:", *qr.splitlines()], sep="\n   ")
            print()
        return self._sparql(qr, limit, offset)

    def _sparql(self, query: str, limit: int, offset: int) -> List[Resource]:
        # POLICY Should notify of failures with exception QueryingError including a message.
        # POLICY Resource _store_metadata should not be set (default is None).
        # POLICY Resource _synchronized should not be set (default is False).
        not_supported()

    # Versioning.

    def freeze(self, data: Union[Resource, List[Resource]]) -> None:
        # Replace None by self._freeze_many to switch to optimized bulk freezing.
        run(self._freeze_one, None, data, id_required=True, required_synchronized=True,
            exception=FreezingError)

    def _freeze_many(self, resources: List[Resource]) -> None:
        # Bulk freezing could be optimized by overriding this method in the specialization.
        # POLICY Should reproduce self._freeze_one() and execution._run_one() behaviours.
        not_supported()

    def _freeze_one(self, resource: Resource) -> None:
        # Notify of failures with exception FreezingError including a message.
        # Use self.versioned_id_template.format(x=resource) to freeze IDs.
        for _, v in resource.__dict__.items():
            if isinstance(v, List):
                for x in v:
                    if isinstance(x, Resource):
                        self._freeze_one(x)
            elif isinstance(v, Resource):
                self._freeze_one(v)
        if hasattr(resource, "id"):
            resource.id = self.versioned_id_template.format(x=resource)

    # Utils.

    @abstractmethod
    def _initialize_service(self, endpoint: Optional[str], bucket: Optional[str],
                            token: Optional[str]) -> Any:
        # POLICY Should initialize the access to the store according to its configuration.
        pass


def rewrite_sparql(query: str, context: Dict[str, Dict], prefixes: Optional[Dict[str, str]]
                   ) -> str:
    """Rewrite local property and type names from Model.template() as IRIs.

    Local names are mapped to IRIs by using a JSON-LD context, i.e. { "@context": { ... }}.
    In the case of contexts using prefixed names, prefixes are added to the SPARQL query prologue.
    """
    ctx = {k: v["@id"] if isinstance(v, Dict) else v for k, v in context["@context"].items()}

    if ctx.get("type") == "@type":
        if "rdf" in prefixes:
            ctx["type"] = "rdf:type"
        else:
            ctx["type"] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

    def replace(match: Match) -> str:
        m4 = match.group(4)
        if m4 is None:
            return match.group(0)
        else:
            v = ctx.get(m4, m4)
            m5 = match.group(5)
            if "//" in v:
                return f"<{v}>{m5}"
            else:
                return f"{v}{m5}"

    g4 = r"([a-zA-Z_]+)"
    g5 = r"([.;]?)"
    g0 = rf"((?<=[\s,[(/|!^])((a|true|false)|{g4}){g5}(?=[\s,\])/|?*+]))"
    g6 = r"(('[^']+')|('''[^\n\r]+''')|(\"[^\"]+\")|(\"\"\"[^\n\r]+\"\"\"))"
    rx = rf"{g0}|{g6}"
    qr = re.sub(rx, replace, query)

    if prefixes is None:
        return qr
    else:
        pfx = "\n".join(f"PREFIX {k}: <{v}>" for k, v in prefixes.items())
        return f"{pfx}\n{qr}"



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

import asyncio
import json
import mimetypes
import re
from asyncio import Semaphore, Task
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import quote_plus, unquote

import nexussdk as nexus
import requests
from aiohttp import ClientSession, MultipartWriter
from aiohttp.hdrs import CONTENT_DISPOSITION, CONTENT_TYPE
from pyld import jsonld
from rdflib import Graph
from rdflib.plugins.sparql.parser import Query
from requests import HTTPError

from kgforge.core import Resource
from kgforge.core.archetypes import Store
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import (DeprecationError, DownloadingError, QueryingError,
                                             RegistrationError, RetrievalError, TaggingError,
                                             UpdatingError, UploadingError)
from kgforge.core.commons.execution import run
from kgforge.core.conversions.rdf import as_jsonld, from_jsonld
from kgforge.specializations.mappers import DictionaryMapper
from kgforge.specializations.mappings import DictionaryMapping
from kgforge.specializations.stores.nexus.service import BatchAction, DEPRECATED_PROPERTY, Service


class CategoryDataType(Enum):
    NUMBER = "number"
    BOOLEAN = "boolean"
    LITERAL = "literal"


type_map = {
    str: CategoryDataType.LITERAL,
    bool:  CategoryDataType.BOOLEAN,
    int: CategoryDataType.NUMBER,
    float: CategoryDataType.NUMBER,
    complex: CategoryDataType.NUMBER
}

format_type = {
    CategoryDataType.NUMBER: lambda x: x,
    CategoryDataType.LITERAL: lambda x: f"\"{x}\"",
    CategoryDataType.BOOLEAN: lambda x: "true" if x is True else "false"
}

operator_map = {
    "__lt__": "<",
    "__le__": "<=",
    "__eq__": "=",
    "__ne__": "!=",
    "__gt__": ">",
    "__ge__": ">=",
}


class BlueBrainNexus(Store):

    def __init__(self, endpoint: Optional[str] = None, bucket: Optional[str] = None,
                 token: Optional[str] = None, versioned_id_template: Optional[str] = None,
                 file_resource_mapping: Optional[str] = None,
                 model_context: Optional[Context] = None) -> None:
        super().__init__(endpoint, bucket, token, versioned_id_template, file_resource_mapping,
                         model_context)

    @property
    def mapping(self) -> Optional[Callable]:
        return DictionaryMapping

    @property
    def mapper(self) -> Optional[Callable]:
        return DictionaryMapper

    def register(self, data: Union[Resource, List[Resource]], schema_id: str = None) -> None:
        run(self._register_one, self._register_many, data, required_synchronized=False,
            execute_actions=True, exception=RegistrationError, monitored_status="_synchronized",
            schema_id=schema_id)

    def _register_many(self, resources: List[Resource], schema_id: str) -> None:

        def register_callback(task: Task):
            result = task.result()
            if isinstance(result.response, Exception):
                self.service.synchronize_resource(
                    result.resource, result.response, self._register_many.__name__, False, False)
            else:
                result.resource.id = result.response["@id"]
                if not hasattr(result.resource, "context"):
                    context = self.model_context or self.context
                    result.resource.context = context.iri if context.is_http_iri() else context.document["@context"]
                self.service.synchronize_resource(
                    result.resource, result.response, self._register_many.__name__, True, True)

        verified = self.service.verify(
            resources, self._register_many.__name__, RegistrationError, id_required=False,
            required_synchronized=False, execute_actions=True)
        self.service.batch_request(
            verified, BatchAction.CREATE, register_callback, RegistrationError, schema_id=schema_id)

    def _register_one(self, resource: Resource, schema_id: str) -> None:
        context = self.model_context or self.context
        data = as_jsonld(resource, "compacted", False, model_context=context,
                         metadata_context=None, context_resolver=self.service.resolve_context)
        try:
            response = nexus.resources.create(org_label=self.organisation,
                                              project_label=self.project, data=data,
                                              schema_id=schema_id)
        except nexus.HTTPError as e:
            raise RegistrationError(_error_message(e))
        else:
            resource.id = response['@id']
            # If resource had no context, update it with the one provided by the store.
            if not hasattr(resource, "context"):
                resource.context = data["@context"]
            self.service.sync_metadata(resource, response)

    def _upload_many(self, paths: List[Path], content_type: str) -> List[Dict]:

        async def _bulk():
            loop = asyncio.get_event_loop()
            semaphore = Semaphore(self.service.max_connections)
            async with ClientSession(headers=self.service.headers_upload) as session:
                tasks = (_create_task(x, loop, semaphore, session) for x in paths)
                return await asyncio.gather(*tasks)

        def _create_task(path, loop, semaphore, session):
            default = "application/octet-stream"
            mime_type = (content_type or mimetypes.guess_type(str(path))[0] or default)
            # FIXME Nexus seems to not parse the Content-Disposition 'filename*' field  properly.
            # data = FormData()
            # data.add_field("file", path.open("rb"), content_type=mime_type, filename=path.name)
            # FIXME This hack is to prevent sending Content-Disposition with the 'filename*' field.
            data = MultipartWriter("form-data")
            part = data.append(path.open("rb"))
            part.headers[CONTENT_TYPE] = mime_type
            part.headers[CONTENT_DISPOSITION] = f'form-data; name="file"; filename="{path.name}"'
            return loop.create_task(_upload(data, semaphore, session))

        async def _upload(data, semaphore, session):
            async with semaphore:
                async with session.post(self.service.url_files, data=data) as response:
                    body = await response.json()
                    if response.status < 400:
                        return body
                    else:
                        msg = " ".join(re.findall('[A-Z][^A-Z]*', body["@type"])).lower()
                        raise UploadingError(msg)

        return asyncio.run(_bulk())

    def _upload_one(self, path: Path, content_type: str) -> Dict:
        file = str(path.absolute())
        mime_type = content_type or mimetypes.guess_type(file, True)[0]
        if mime_type is None:
            mime_type = "application/octet-stream"
        try:
            response = nexus.files.create(self.organisation, self.project, file,
                                          content_type=mime_type)
        except HTTPError as e:
            raise UploadingError(_error_message(e))
        else:
            return response

    # C[R]UD.

    def retrieve(self, id: str, version: Optional[Union[int, str]],
                 cross_bucket: bool) -> Resource:
        segment = "resolvers" if cross_bucket else "resources"
        url = f"{self.endpoint}/{segment}/{self.bucket}/_/{quote_plus(id)}"
        if version is not None:
            if isinstance(version, int):
                params = {"rev": version}
            elif isinstance(version, str):
                params = {"tag": version}
            else:
                raise RetrievalError("incorrect 'version'")
        else:
            params = None
        try:
            response = requests.get(url, params=params, headers=self.service.headers)
            response.raise_for_status()
        except HTTPError as e:
            raise RetrievalError(_error_message(e))
        else:
            data = response.json()
            resource = self.service.to_resource(data)
            resource._synchronized = True
            self.service.sync_metadata(resource, data)
            return resource

    def _retrieve_filename(self, id: str) -> str:
        try:
            response = requests.get(id, headers=self.service.headers)
            response.raise_for_status()
            metadata = response.json()
            return metadata["_filename"]
        except HTTPError as e:
            raise DownloadingError(_error_message(e))

    def _download_many(self, urls: List[str], paths: List[str]) -> None:

        async def _bulk():
            loop = asyncio.get_event_loop()
            semaphore = Semaphore(self.service.max_connections)
            async with ClientSession(headers=self.service.headers_download) as session:
                tasks = (_create_task(x, y, loop, semaphore, session) for x, y in zip(urls, paths))
                return await asyncio.gather(*tasks)

        def _create_task(url, path, loop, semaphore, session):
            return loop.create_task(_download(url, path, semaphore, session))

        async def _download(url, path, semaphore, session):
            async with semaphore:
                async with session.get(url) as response:
                    if response.status < 400:
                        with open(path, "wb") as f:
                            data = await response.read()
                            f.write(data)
                    else:
                        error = await response.json()
                        msg = " ".join(re.findall('[A-Z][^A-Z]*', error["@type"])).lower()
                        raise DownloadingError(msg)

        return asyncio.run(_bulk())

    def _download_one(self, url: str, path: str) -> None:
        try:
            # this is a hack since _self and _id have the same uuid
            file_id = url.split("/")[-1]
            if file_id.startswith("http"):
                file_id = unquote(file_id).split("/")[-1]
            if len(file_id) < 1:
                raise DownloadingError("Invalid file name")
            nexus.files.fetch(org_label=self.organisation, project_label=self.project,
                              file_id=file_id, out_filepath=path)
        except HTTPError as e:
            raise DownloadingError(_error_message(e))

    # CR[U]D.

    def update(self, data: Union[Resource, List[Resource]]) -> None:
        run(self._update_one, self._update_many, data, id_required=True, required_synchronized=False,
            execute_actions=True, exception=UpdatingError, monitored_status="_synchronized")

    def _update_many(self, resources: List[Resource]) -> None:
        update_callback = self.service.default_callback(self._update_many.__name__)
        verified = self.service.verify(
            resources, self._update_many.__name__, UpdatingError, id_required=True,
            required_synchronized=False, execute_actions=True)
        self.service.batch_request(verified, BatchAction.UPDATE, update_callback, UpdatingError)

    def _update_one(self, resource: Resource) -> None:
        context = self.model_context or self.context
        data = as_jsonld(resource, "compacted", False, model_context=context, metadata_context=None,
                         context_resolver=self.service.resolve_context)
        rev = {"rev": resource._store_metadata._rev}
        url = f"{self.service.url_resources}/_/{quote_plus(resource.id)}"
        try:
            response = requests.put(url, headers=self.service.headers,
                                    data=json.dumps(data, ensure_ascii=True), params=rev)
            response.raise_for_status()
        except HTTPError as e:
            raise UpdatingError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

    def tag(self, data: Union[Resource, List[Resource]], value: str) -> None:
        run(self._tag_one, self._tag_many, data, id_required=True, required_synchronized=True,
            exception=TaggingError, value=value)

    def _tag_many(self, resources: List[Resource], value: str) -> None:
        tag_callback = self.service.default_callback(self._tag_many.__name__)
        verified = self.service.verify(
            resources, self._tag_many.__name__, TaggingError, id_required=True,
            required_synchronized=True, execute_actions=False)
        self.service.batch_request(
            verified, BatchAction.TAG, tag_callback, TaggingError, tag=value)

    def _tag_one(self, resource: Resource, value: str) -> None:
        data = {
            "tag": value,
            "rev": resource._store_metadata._rev
        }
        url = f"{self.service.url_resources}/_/{ quote_plus(resource.id)}/tags?rev={resource._store_metadata._rev}"
        try:
            response = requests.post(url, headers=self.service.headers,
                                     data=json.dumps(data, ensure_ascii=True))
            response.raise_for_status()
        except HTTPError as e:
            raise TaggingError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

    # CRU[D].

    def deprecate(self, data: Union[Resource, List[Resource]]) -> None:
        run(self._deprecate_one, self._deprecate_many, data, id_required=True,
            required_synchronized=True, exception=DeprecationError, monitored_status="_synchronized")

    def _deprecate_many(self, resources: List[Resource]) -> None:
        deprecate_callback = self.service.default_callback(self._deprecate_many.__name__)
        verified = self.service.verify(
            resources, self._deprecate_many.__name__, DeprecationError, id_required=True,
            required_synchronized=True, execute_actions=False)
        self.service.batch_request(
            verified, BatchAction.DEPRECATE, deprecate_callback, DeprecationError)

    def _deprecate_one(self, resource: Resource) -> None:
        url = f"{self.service.url_resources}/_/{ quote_plus(resource.id)}?rev={resource._store_metadata._rev}"
        try:
            response = requests.delete(url, headers=self.service.headers)
            response.raise_for_status()
        except HTTPError as e:
            raise DeprecationError(_error_message(e))
        else:
            self.service.sync_metadata(resource, response.json())

    # Querying.

    def search(self, resolvers: Optional[List["Resolver"]], *filters, **params) -> List[Resource]:

        if self.model_context is None:
            raise ValueError("context model missing")

        debug = params.get("debug", False)
        limit = params.get("limit", 100)
        offset = params.get("offset", None)
        deprecated = params.get("deprecated", False)
        query_statements, query_filters = build_query_statements(self.model_context, filters)
        query_statements.insert(0, f"<{DEPRECATED_PROPERTY}> {format_type[CategoryDataType.BOOLEAN](deprecated)}")
        statements = "\n".join((";\n ".join(query_statements), ".\n ".join(query_filters)))

        query = f"SELECT ?id WHERE {{ ?id {statements}}}"
        resources = self.sparql(query, debug=debug, limit=limit, offset=offset)
        results = self.service.batch_request(resources, BatchAction.FETCH, None, QueryingError)
        resources = list()
        for result in results:
            resource = result.resource
            try:
                resource = self.service.to_resource(result.response)
            except Exception as e:
                self.service.synchronize_resource(
                    resource, result.response, self.search.__name__, False, False)
                raise ValueError(e)
            finally:
                self.service.synchronize_resource(
                    resource, result.response, self.search.__name__, True, True)
            resources.append(resource)
        return resources

    def _sparql(self, query: str, limit: int, offset: int = None) -> List[Resource]:

        s_offset = "" if offset is None else f"OFFSET {offset}"
        s_limit = "" if limit is None else f"LIMIT {limit}"
        query = f"{query} {s_limit} {s_offset}"

        try:
            response = requests.post(
                self.service.sparql_endpoint, data=query, headers=self.service.headers_sparql)
            response.raise_for_status()
        except Exception as e:
            raise QueryingError(e)
        else:
            data = response.json()
            # FIXME workaround to parse a CONSTRUCT query, this fix depends on
            #  https://github.com/BlueBrain/nexus/issues/1155
            _, q_comp = Query.parseString(query)
            if q_comp.name == "ConstructQuery":
                subject_triples = {}
                for r in data["results"]["bindings"]:
                    subject = r['subject']['value']
                    s = f"<{r['subject']['value']}>"
                    p = f"<{r['predicate']['value']}>"
                    if r["object"]["type"] == "uri":
                        o = f"<{r['object']['value']}>"
                    else:
                        if "datatype" in r["object"]:
                            o = f"\"{r['object']['value']}\"^^{r['object']['datatype']}"
                        else:
                            o = f"\"{r['object']['value']}\""
                    if subject in subject_triples:
                        subject_triples[subject] += f"\n{s} {p} {o} . "
                    else:
                        subject_triples[subject] = f"{s} {p} {o} . "

                def triples_to_resource(iri, triples):
                    graph = Graph().parse(data=triples, format="nt")
                    data_expanded = json.loads(graph.serialize(format="json-ld").decode("utf-8"))
                    frame = {"@id": iri}
                    data_framed = jsonld.frame(data_expanded, frame)
                    context = self.model_context or self.context
                    compacted = jsonld.compact(data_framed, context.document)
                    resource = from_jsonld(compacted)
                    resource.context = context.iri if context.is_http_iri() else context.document["@context"]
                    return resource

                return [triples_to_resource(s, t) for s, t in subject_triples.items()]

            else:
                # SELECT QUERY
                results = data["results"]["bindings"]
                return [Resource(**{k: v["value"] for k, v in x.items()}) for x in results]

    # Utils.

    def _initialize_service(self, endpoint: Optional[str], bucket: Optional[str],
                            token: Optional[str]) -> Any:
        try:
            self.organisation, self.project = self.bucket.split('/')
        except ValueError:
            raise ValueError("malformed bucket parameter, expecting 'organization/project' like")
        else:
            return Service(endpoint, self.organisation, self.project, token, self.model_context, 200)


def _error_message(error: HTTPError) -> str:

    def format_message(msg):
        return "".join([msg[0].lower(), msg[1:-1], msg[-1] if msg[-1] != "." else ""])

    try:
        reason = error.response.json()["reason"]
        return format_message(reason)
    except AttributeError:
        pass
    try:
        return format_message(error.response.text())
    except AttributeError:
        return format_message(str(error))


def build_query_statements(context: Context, *conditions) -> Tuple[List,List]:
    statements = list()
    filters = list()
    for index, f in enumerate(*conditions):
        try:
            last_term = context.terms[f.path[-1]]
        except KeyError:
            last_term = ""

        if f.path[-1] == "id":
            f.path.pop()

        property_path = "/".join(f.path)

        if f.path[-1] == "type" or last_term.type == "@id":
            if f.operator == "__eq__":
                statements.append(f"{property_path} {f.value}")
            elif f.operator == "__ne__":
                statements.append(f"{property_path} ?v{index}")
                filters.append(f"FILTER(?v{index} != {f.value})")
            else:
                raise NotImplementedError(f"operator '{f.operator}' is not supported in this query")
        else:
            value_type = type_map[type(f.value)]
            value = format_type[value_type](f.value)
            if value_type is CategoryDataType.LITERAL:
                statements.append(f"{property_path} ?v{index}")
                filters.append(f"FILTER(?v{index} = {value})")
                # filters.append(f"{property_path} ?v{index} FILTER regex(?v{index}, {value})")
            else:
                statements.append(f"{property_path} ?v{index}")
                filters.append(f"FILTER(?v{index} {operator_map[f.operator]} {value})")

    return statements, filters




# 
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.


# Forge operations.

class ConfigurationError(Exception):
    pass


class NotSupportedError(Exception):
    pass


# Model operations.


class ValidationError(Exception):
    pass


# Resolver operations.


class ResolvingError(Exception):
    pass


# Store operations.


class RegistrationError(Exception):
    pass


class UploadingError(Exception):
    pass


class RetrievalError(Exception):
    pass


class DownloadingError(Exception):
    pass


class UpdatingError(Exception):
    pass


class TaggingError(Exception):
    pass


class DeprecationError(Exception):
    pass


class QueryingError(Exception):
    pass


class FreezingError(Exception):
    pass



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.
import types
from abc import abstractmethod
from typing import List, Dict, Tuple, Set, Optional
from pyshacl.shape import Shape
from pyshacl.shapes_graph import ShapesGraph
from rdflib import Graph, URIRef, RDF, XSD

from kgforge.core import Resource
from kgforge.core.commons.context import Context
from kgforge.core.commons.exceptions import ConfigurationError
from kgforge.core.conversions.rdf import as_graph
from kgforge.specializations.models.rdf.collectors import (AndCollector, NodeCollector,
                                                           PropertyCollector, MinCountCollector,
                                                           DatatypeCollector, InCollector,
                                                           ClassCollector, NodeKindCollector,
                                                           OrCollector, XoneCollector,
                                                           HasValueCollector)
from kgforge.specializations.models.rdf.node_properties import NodeProperties
from kgforge.specializations.models.rdf.utils import as_term

ALL_COLLECTORS = [
    AndCollector,
    OrCollector,
    PropertyCollector,
    NodeCollector,
    PropertyCollector,
    MinCountCollector,
    DatatypeCollector,
    InCollector,
    ClassCollector,
    NodeKindCollector,
    XoneCollector,
    HasValueCollector
]
ALL_COLLECTORS_MAP = {c.constraint(): c for c in ALL_COLLECTORS}


def traverse(self, predecessors: Set[URIRef]) -> Tuple[List, Dict]:
    """ traverses the Shape SACL properties to collect constrained properties

    This function is injected to pyshacl Shape object in order to traverse the Shacl graph.
    It will call a specific collector depending on the SHACL property present in the NodeShape

    Args:
        predecessors: list of nodes that have being traversed, used to break circular
            recursion

    Returns:
        properties, attributes: Tuple(list,dict), the collected properties and attributes
            respectively gathered from the collectors
    """

    parameters = self.parameters()
    properties = list()
    attributes = dict()
    done_collectors = set()
    for param in iter(parameters):
        if param in ALL_COLLECTORS_MAP:
            constraint_collector = ALL_COLLECTORS_MAP[param]
            if constraint_collector not in done_collectors:
                c = constraint_collector(self)
                predecessors.add(self.node)
                props, attrs = c.collect(predecessors)
                if attrs:
                    attributes.update(attrs)
                if props:
                    properties.extend(props)
                done_collectors.add(constraint_collector)
                if predecessors:
                    predecessors.remove(self.node)
        else:
            # FIXME: there are some SHACL constrains that are not implemented
            # raise IndexError(f"{param} not implemented!")
            pass

    return properties, attributes


class ShapesGraphWrapper(ShapesGraph):

    def __init__(self, graph: Graph) -> None:
        super().__init__(graph)
        # the following line triggers the shape loading
        self._shapes = self.shapes

    def lookup_shape_from_node(self, node: URIRef) -> Shape:
        """ Overwrite function to inject the transverse function for only to requested nodes.

        Args:
            node (URIRef): The node to look up.

        Returns:
            Shape: The Shacl shape of the requested node.
        """
        shape = self._node_shape_cache[node]
        if not hasattr(shape, "traverse"):
            shape.traverse = types.MethodType(traverse, shape)
        return shape


class RdfService:

    def __init__(self, graph: Graph, context_iri: Optional[str] = None) -> None:

        if context_iri is None:
            raise ConfigurationError(f"RdfModel requires a context")
        self._graph = graph
        self._context_cache = dict()
        self.classes_to_shapes = self._build_shapes_map()
        resolved_context = self.resolve_context(context_iri)
        self.context = Context(resolved_context, context_iri)
        self.types_to_shapes: Dict = self._build_types_to_shapes()

    def schema_source_id(self, schema_iri: str) -> str:
        # POLICY Should return the id of the resource containing the schema
        raise NotImplementedError()

    @abstractmethod
    def materialize(self, iri: URIRef) -> NodeProperties:
        """Triggers the collection of properties of a given Shape node

        Args:
            iri: the URI of the node to start collection

        Returns:
            A NodeProperty object with the collected properties
        """
        raise NotImplementedError()

    def validate(self, resource: Resource):
        try:
            shape_iri = self.types_to_shapes[resource.type]
        except AttributeError:
            raise TypeError("resource requires a type attribute")
        else:
            data_graph = as_graph(resource, False, self.context, None, None)
            return self._validate(shape_iri, data_graph)

    @abstractmethod
    def _validate(self, iri: str, data_graph: Graph) -> Tuple[bool, Graph, str]:
        raise NotImplementedError()

    @abstractmethod
    def resolve_context(self, iri: str) -> Dict:
        """For a given IRI return its resolved context recursively"""
        raise NotImplementedError()

    @abstractmethod
    def generate_context(self) -> Dict:
        """Generates a JSON-LD context with the classes and terms present in the SHACL graph."""
        raise NotImplementedError()

    @abstractmethod
    def _build_shapes_map(self) -> Dict:
        """Queries the source and returns a map of owl:Class to sh:NodeShape"""
        raise NotImplementedError()

    def _build_types_to_shapes(self):
        """Iterates the classes_to_shapes dictionary to create a term to shape dictionary filtering
         the terms available in the context """
        types_to_shapes: Dict = dict()
        for k, v in self.classes_to_shapes.items():
            term = self.context.find_term(str(k))
            if term:
                key = term.name
                if term.name not in types_to_shapes:
                    types_to_shapes[term.name] = v
                else:
                    print("WARN: duplicated term", key, k, [key], v)
        return types_to_shapes

    def _generate_context(self) -> Dict:
        """Materializes all Types into templates and parses the templates to generate a context"""
        # FIXME: the status of this function is experimental
        # TODO: check if there are conflicting terms, and throw error
        context = dict()
        prefixes = dict()
        types_ = dict()
        terms = dict()

        def traverse_properties(properties) -> Tuple[Dict, Dict]:
            l_prefixes = dict()
            l_terms = dict()
            for property_ in properties:
                if hasattr(property_, "path"):
                    if property_.path != RDF.type and str(property_.path) != "id":
                        v_prefix, v_namespace, v_name = self._graph.compute_qname(property_.path)
                        l_prefixes.update({v_prefix: str(v_namespace)})
                        term_obj = {"@id": ":".join((v_prefix, v_name))}
                        if hasattr(property_, "id"):
                            term_obj.update({"@type": "@id"})
                        if hasattr(property_, "values"):
                            if isinstance(property_.values, str) or len(property_.values) == 1:
                                if isinstance(property_.values, list):
                                    obj_type = property_.values[0]
                                else:
                                    obj_type = property_.values
                                if obj_type in target_classes:
                                    term_obj.update({"@type": "@id"})
                                else:
                                    try:
                                        px, ns, n = self.graph.compute_qname(obj_type)
                                        l_prefixes.update({px: str(ns)})
                                        if str(ns) == str(XSD):
                                            term_obj.update({"@type": ":".join((px, n))})
                                    except Exception:
                                        pass
                        l_terms.update({v_name: term_obj})
                if hasattr(property_, "properties"):
                    l_p, l_t = traverse_properties(property_.properties)
                    l_prefixes.update(l_p)
                    l_terms.update(l_t)
            return l_prefixes, l_terms

        target_classes = list()
        for k in self.classes_to_shapes.keys():
            key = as_term(k)
            if key not in target_classes:
                target_classes.append(key)
            else:
                # TODO: should this raise an error?
                print("duplicated term", key, k)

        for type_, shape in self.classes_to_shapes.items():
            t_prefix, t_namespace, t_name = self._graph.compute_qname(type_)
            prefixes.update({t_prefix: str(t_namespace)})
            types_.update({t_name: {"@id": ":".join((t_prefix, t_name))}})
            node = self.materialize(shape)
            if hasattr(node, "properties"):
                p, t = traverse_properties(node.properties)
                prefixes.update(p)
                terms.update(t)

        context.update({key: prefixes[key] for key in sorted(prefixes)})
        context.update({key: types_[key] for key in sorted(types_)})
        context.update({key: terms[key] for key in sorted(terms)})

        return {"@context":  context} if len(context) > 0 else None



#
# Blue Brain Nexus Forge is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Blue Brain Nexus Forge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Blue Brain Nexus Forge. If not, see <https://choosealicense.com/licenses/lgpl-3.0/>.

from typing import Any, Dict, Optional, Union

import hjson

from kgforge.core.commons.attributes import check_collisions, repr_class, sort_attrs
from kgforge.core.wrappings.dict import DictWrapper


# NB: Do not 'from kgforge.core.commons.actions import Action' to avoid cyclic dependency.


class Resource:

    # See datasets.py in kgforge/specializations/resources/ for a reference specialization.

    # POLICY Specializations could add methods and attributes if they are added to _RESERVED.
    # TODO Move from BDD to classical testing to have a more parameterizable test suite. DKE-135.
    # POLICY Specializations should pass tests/core/resources/resource.feature tests.

    # FIXME remove "_forge" from _RESERVED and implement a "cast" of derived classes to Resource
    _RESERVED = {"_last_action", "_validated", "_synchronized", "_store_metadata", "_forge"}

    def __init__(self, **properties) -> None:
        check_collisions(self._RESERVED, properties.keys())
        self.__dict__.update(properties)
        # Status of the last modifying action performed on the resource.
        self._last_action: Optional["Action"] = None
        # True if the resource has been validated.
        # False if the resource has not been validated yet or a modification has been done since
        # the last validation.
        self._validated: bool = False
        # True if the resource is synchronized with the store.
        # False if the resource has not been registered yet or a modification has been done since
        # the synchronization.
        self._synchronized: bool = False
        # None until synchronized.
        # Otherwise, holds the metadata the store returns at synchronization.
        self._store_metadata: Optional[DictWrapper] = None

    def __repr__(self) -> str:
        return repr_class(self)

    def __str__(self) -> str:
        return hjson.dumps(self, indent=4, default=encode, item_sort_key=sort_attrs)

    def __eq__(self, other: object) -> bool:
        def _data(resource: Resource) -> Dict:
            return {k: _data(v) if isinstance(v, Resource) else v
                    for k, v in resource.__dict__.items()
                    if k not in resource._RESERVED}
        if type(other) is type(self):
            sdict = _data(self)
            odict = _data(other)
            return sdict == odict
        else:
            return False

    def __setattr__(self, key, value) -> None:
        if key not in self._RESERVED:
            self.__dict__["_validated"] = False
            self.__dict__["_synchronized"] = False
        self.__dict__[key] = value


def encode(data: Any) -> Union[str, Dict]:
    if isinstance(data, Resource):
        return {k: v for k, v in data.__dict__.items() if k not in data._RESERVED}
    elif type(data).__name__ == "LazyAction":
        return str(data)
    else:
        return data.__dict__
